# -*- coding: utf-8 -*-
"""
Created on Wed Mar  2 08:22:48 2022

@author: xavier.mouy
"""

# from ecosound.core.annotation import Annotation
# from ecosound.core.spectrogram import Spectrogram
from ecosound.core.measurement import Measurement

# from ecosound.core.audiotools import Sound
# from ecosound.visualization.grapher_builder import GrapherFactory


dataset_file_path = r"D:\NOAA\2022_BC_fish_detector\manual_annotations\Master_annotations_dataset_20221028_without_06-MILL-FS.nc"
out_dir = r"D:\NOAA\2022_BC_fish_detector\spectrograms\SNR_test2"


# Load dataset
dataset = Measurement()
dataset.from_netcdf(dataset_file_path)
# dataset.filter("deployment_ID=='SI-RCAOut-20181015'", inplace=True)
# dataset.filter("deployment_ID=='SI-RCAOut-20181015'", inplace=True)
dataset.filter("label_class=='FS'", inplace=True)
# dataset.data = dataset.data[0:10]

# dataset.update_audio_dir(
#     r"D:\NOAA\2022_BC_fish_detector\manual_annotations", verbose=True
# )

dataset.export_spectrograms(
    out_dir,
    time_buffer_sec=0.2,
    spectro_unit="sec",
    spetro_nfft=0.064,
    spetro_frame=0.064,
    spetro_inc=0.00125,
    freq_min_hz=None,
    freq_max_hz=None,
    sanpling_rate_hz=4000,
    filter_order=8,
    filter_type="iir",
    fig_size=(15, 10),
    deployment_subfolders=False,
    date_subfolders=False,
    # file_name_field="time_min_offset",
    # file_name_field="uuid",
    file_name_field="uuid",
    file_prefix_field="snr",
    channel=0,
    # colormap="viridis",
    colormap="Greys",
    save_wav=True,
)



# -*- coding: utf-8 -*-
"""

@author: xavier.mouy
"""

from ecosound.core.tools import list_files
import re

from ecosound.core.annotation import Annotation
from ecosound.core.audiotools import Sound
import soundfile as sf
from datetime import datetime
import matplotlib.pyplot as plt
import numpy as np
from sklearn.model_selection import StratifiedGroupKFold
import scipy
import os
import csv


annot_dataset_file = (
    r"D:\Detector\datasets\Master_annotations_dataset_20221025.nc"
)

new_data_dir = r"D:\Detector\datasets2"
new_dataset_file = (
    r"D:\Detector\datasets\Master_annotations_dataset_2022-10-25_test.nc"
)

# Load dataset
dataset = Annotation()
dataset.from_netcdf(annot_dataset_file)


# update audio dir path
dataset.update_audio_dir(new_data_dir, verbose=False)

# # list name of all audio files in dataset
# dataset_files_list = set(
#     dataset.data["audio_file_dir"]
#     + os.path.sep
#     + dataset.data["audio_file_name"]
#     + dataset.data["audio_file_extension"]
# )

# # list extension of all audio files in dataset
# dataset_ext_list = set(
#     [os.path.splitext(file)[1] for file in dataset_files_list]
# )

# # list all audio files in new folder (only for the target file extensions)
# new_dir_files_list = []
# for ext in dataset_ext_list:
#     new_dir_files_list = new_dir_files_list + list_files(
#         new_data_dir, ext, recursive=True
#     )

# # go through each file in dataset and try to find in in new data folder
# missing_files_list = []
# for file in dataset_files_list:
#     res = [
#         idx
#         for idx, new_dir_file in enumerate(new_dir_files_list)
#         if re.search(os.path.split(file)[1], new_dir_file)
#     ]
#     if len(res) == 0:
#         missing_files_list.append(file)
#     else:
#         new_path = os.path.split(new_dir_files_list[res[0]])[0]
#         dataset.data.loc[
#             dataset.data["audio_file_name"]
#             == os.path.splitext(os.path.split(file)[1])[0],
#             "audio_file_dir",
#         ] = new_path

# if len(missing_files_list) > 0:
#     print(str(len(missing_files_list)), " files could not be found.")
#     print(missing_files_list)

# # save update dataset
# dataset.to_netcdf(new_dataset_file)



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 19 10:16:31 2022

@author: xavier.mouy
"""
from ecosound.core.annotation import Annotation
from ecosound.evaluation.prf import PRF
import numpy as np

###############################################################################
##  input parameters ##########################################################
###############################################################################

# # Annotation file:
# annot_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20201102\Annotations_dataset_UK-SAMS-WestScotland-202011-N1 annotations_withSNR.nc"
# # Detection folder or detection sqllite file
# detec_file = r"C:\Users\xavier.mouy\Documents\GitHub\ketos_utils\results\UK-SAMS-N1\spectro-5s_fft-0.128_step-0.064_fmin-0_fmax-800_no-norm_models-5s_bs128_ep50\detections.sqlite"
# # output folder
# out_dir = r"C:\Users\xavier.mouy\Documents\GitHub\ketos_utils\results\UK-SAMS-N1\spectro-5s_fft-0.128_step-0.064_fmin-0_fmax-800_no-norm_models-5s_bs128_ep50\performance_results"

# Annotation file:
annot_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20201102\Annotations_dataset_UK-SAMS-WestScotland-202011-N1 annotations_withSNR.nc"
# Detection folder or detection sqllite file
detec_file = r"C:\Users\xavier.mouy\Documents\GitHub\ketos_utils\results\UK-SAMS-N1\spectro-5s_fft-0.256_step-0.03_fmin-0_fmax-800_no-norm_models-5s_bs128_ep50\detections.sqlite"
# output folder
out_dir = r"C:\Users\xavier.mouy\Documents\GitHub\ketos_utils\results\UK-SAMS-N1\spectro-5s_fft-0.256_step-0.03_fmin-0_fmax-800_no-norm_models-5s_bs128_ep50\performance_results"


# Detection threshold to test
thresholds = np.arange(0, 1, 0.01)
# thresholds = [0.9]

# Name of the class to test
target_class = "MW"

# List of files to use
files_to_use = "detec"  # 'detec', 'annot', 'both', list

# restrictions on min and max dates:
date_min = None  # date string or None
date_max = None  # date string or None

# Beta parameter for calculating the F-score
F_beta = 1

# Matching criteria
freq_ovp = True
dur_factor_max = None
dur_factor_min = None
ovlp_ratio_min = None
remove_duplicates = False
inherit_metadata = False
filter_deploymentID = False
do_plot = False

###############################################################################
###############################################################################

# load ground truth data
print(" ")
print("Loading manual annotations...")
annot = Annotation()
annot.from_netcdf(annot_file)
print("Annotation labels:")
print(annot.get_labels_class())

# load destections
print(" ")
print("Loading automatic detections...")
detec = Annotation()
detec.from_sqlite(detec_file)

# calculate performance
print(" ")
print("Calculating performance...")
PRF.count(
    annot=annot,
    detec=detec,
    out_dir=out_dir,
    target_class=target_class,
    files_to_use=files_to_use,
    date_min=date_min,
    date_max=date_max,
    thresholds=thresholds,
    F_beta=F_beta,
    freq_ovp=freq_ovp,
    dur_factor_max=dur_factor_max,
    dur_factor_min=dur_factor_min,
    ovlp_ratio_min=ovlp_ratio_min,
    remove_duplicates=remove_duplicates,
    inherit_metadata=inherit_metadata,
    filter_deploymentID=filter_deploymentID,
    do_plot=do_plot,
)



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 13:35:34 2022

@author: xavier.mouy
"""

import sys
sys.path.append(r"C:\Users\xavier.mouy\Documents\GitHub\ecosound") # Adds higher directory to python modules path.
from ecosound.core.annotation import Annotation
import ecosound

infile = r'C:\Users\xavier.mouy\Documents\GitHub\ecosound\data\sqlite_annotations/read/detections1.sqlite'

# load sqlite tables
annot = Annotation()
annot.from_sqlite(infile, verbose=True)

# from ecosound.visualization.grapher_builder import GrapherFactory
# graph = GrapherFactory('AnnotHeatmap')
# graph.add_data(annot)
# graph.add_data(annot)
# graph.norm_value=[None, 20]
# graph.title='My title'
# graph.integration_time='60Min'
# graph.colormap='viridis'
# graph.is_binary=False
# graph.date_format = '%d-%b-%Y' #'%d-%b-%Y'
# graph.show()

#graph.to_file(r'C:\Users\xavier.mouy\Desktop\2Daggregate')

annot.heatmap(title='My title Yeah', norm_value=10)



 


# -*- coding: utf-8 -*-
"""
Created on Tue May 23 14:22:20 2023

@author: xavier.mouy
"""
from ecosound.core.measurement import Measurement


infile = r'C:\Users\xavier.mouy\Documents\Projects\2022_DFO_fish_catalog\Darienne_data\Taylor-Islet_LA_dep2\results\AMAR173.1.20220821T180710Z.wav.nc'
outdir = r'C:\Users\xavier.mouy\Documents\Projects\2022_DFO_fish_catalog\Darienne_data\Taylor-Islet_LA_dep2\results'

loc = Measurement()
loc.from_netcdf(infile)

loc.to_raven(outdir)


from ecosound.core.measurement import Measurement
import os
import matplotlib
import matplotlib.pyplot as plt
matplotlib.use('TkAgg')

in_dir =r'C:\Users\xavier.mouy\Documents\Projects\2022_DFO_fish_catalog\Darienne_data\DangerRocksFiles-20230125T183142Z-001\results'
min_err_val = 0.2 # minimum localization uncertainty (in m)

# load data
meas = Measurement()
meas.from_netcdf(in_dir)

# filter based on uncertainty
meas.filter('x_err_span_m < @min_err_val',min_err_val=min_err_val,inplace=True)
meas.filter('y_err_span_m < @min_err_val',min_err_val=min_err_val,inplace=True)
meas.filter('z_err_span_m < @min_err_val',min_err_val=min_err_val,inplace=True)

# plot
X=meas.data.x_m
Y=meas.data.y_m
Z=meas.data.z_m


fig = plt.figure()
ax = fig.add_subplot(121,projection='3d')
ax.scatter(X, Y, Z, c='k',marker='.', alpha=0.5, s=5)
ax.set_xlabel('X (m)')
ax.set_ylabel('Y (m)')
ax.set_zlabel('Z (m)')
ax.set_aspect('equal')
ax.grid(True)
#ax.set_title('Localizations (uncertainty < '+str(min_err_val))

ax = fig.add_subplot(333)
ax.scatter(X, Y, c='k',marker='.', alpha=0.5, s=4)
ax.set_xlabel('X (m)')
ax.set_ylabel('Y (m)')
#ax.set_aspect('equal', 'box')
#ax.grid(True)

ax = fig.add_subplot(336)
ax.scatter(X, Z, c='k',marker='.', alpha=0.5, s=4)
ax.set_xlabel('X (m)')
ax.set_ylabel('Z (m)')
#ax.set_aspect('equal','box')
#ax.grid(True)

ax = fig.add_subplot(339)
ax.scatter(Y, Z, c='k',marker='.', alpha=0.5, s=4)
ax.set_xlabel('Y (m)')
ax.set_ylabel('Z (m)')
#ax.set_aspect('equal','box')
#ax.grid(True)

fig.suptitle('Localizations (uncertainty < '+str(min_err_val)+'m)', fontsize=16)

plt.show()
print('done')




# -*- coding: utf-8 -*-
"""
Created on Fri Feb  7 15:27:39 2020

@author: xavier.mouy
"""

from .measurer_builder import BaseClass
from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.core.audiotools import Sound
import numpy as np
import pandas as pd
from dask import delayed, compute, visualize
import os


class SNR(BaseClass):
    """ """

    measurer_parameters = ("noise_win_sec",)

    def __init__(self, *args, **kwargs):
        """
        Initialize the measurer.

        Parameters
        ----------
        *args : str
            Do not use. Only used by the MeasurerFactory.
        noise_win_sec : float, optional
            Duration in seconds of the window to use to estimate noise. Half of this window is used before and half if
            used after teh signal of interest.

        Returns
        -------
        None. Measurer object.

        """
        # Initialize all measurer parameters to None
        self.__dict__.update(
            dict(
                zip(
                    self.measurer_parameters,
                    [None] * len(self.measurer_parameters),
                )
            )
        )

        # Unpack kwargs as measurer parameters if provided on instantiation
        self.__dict__.update(**kwargs)

    @property
    def name(self):
        """Return name of the measurer."""
        measurer_name = "SNR"
        return measurer_name

    @property
    def version(self):
        """Return version of the measurer."""
        version = "0.1"
        return version

    def _prerun_check(self, annotations):
        """Run several verifications before the run."""
        # check that all required arguments are defined
        if True in [
            self.__dict__.get(keys) is None
            for keys in self.measurer_parameters
        ]:
            raise ValueError(
                "Not all measurer parameters have been defined."
                + " Required parameters: "
                + str(self.measurer_parameters)
            )
        # check that annotations is an Annotation class
        if not isinstance(annotations, Annotation):
            raise ValueError(
                "Input must be an ecosound Annotation object"
                + "(ecosound.core.annotation)."
            )

    def compute(self, annotations, debug=False, verbose=False, use_dask=False):
        """Compute signal-to-noise-ratio of annotations.

        Goes through each annotation and computes the SNR by estinating the
        power of the noise before and after the annotation. Measurements are
        performed on the band-pass filtered waveform.

        Parameters
        ----------
        annotations : ecosound Annotation object
            Annotations of the sounds to measure. Can be from manual analysis
            or from an automatic detector.
        use_dask : bool, optional
            If True, run the measurer in parallele using Dask. The default is
            False.
        debug : bool, optional
            Displays figures for each annotation with the spectrogram, spectral
            and time envelopes, and tables with all associated measurements.
            The default is False.
        verbose : bool, optional
            Prints in the console the annotation being processed. The default
            is False.

        Returns
        -------
        measurements : ecosound Measurement object
            Measurement object containing the measurements appended to the
            original annotation fields. Measurements are in the .data data
            frame. Metadata with mearurer name, version and measurements names
            are in the .metadata datafreame.

        """
        self._prerun_check(annotations)

        # init
        features = self._init_dataframe()
        features_name = list(features.columns)
        # loop through each annotation
        df_list = []
        for index, annot in annotations.data.iterrows():
            if verbose:
                print(
                    "processing annotation ",
                    index,
                    annot["time_min_offset"],
                    "-",
                    annot["time_max_offset"],
                )

            # feature for 1 annot
            if use_dask:
                df = delayed(self.compute_single_annot)(annot, debug)
            else:
                df = self.compute_single_annot(annot, debug)
            # stack features for each annotation
            df_list.append(df)
        if use_dask:
            features = delayed(pd.concat)(df_list, ignore_index=False)
            # features.visualize('measuremnets')
            features = features.compute()
        else:
            features = pd.concat(df_list, ignore_index=False)
        # merge with annotation fields
        annotations.data.set_index("uuid", inplace=True, drop=False)
        features.set_index("uuid", inplace=True, drop=True)
        meas = pd.concat([annotations.data, features], axis=1, join="inner")
        meas.reset_index(drop=True, inplace=True)

        params_dict = dict()
        for param in self.measurer_parameters:
            params_dict[param] = eval("self." + param)

        # create Measurement object
        measurements = Measurement(
            measurer_name=self.name,
            measurer_version=self.version,
            measurements_name=features_name,
            measurements_parameters=params_dict,
        )
        measurements.data = meas
        return measurements

    def _init_dataframe(self):
        tmp = pd.DataFrame(
            {
                "snr": [],
            }
        )
        return tmp

    def compute_single_annot(self, annot, debug):
        # load sound file properties
        sound = Sound(
            os.path.join(annot["audio_file_dir"], annot["audio_file_name"])
            + annot["audio_file_extension"]
        )

        # verify that time boundaries are correct and fit into the duration of the sound file
        if annot["time_max_offset"] > sound.file_duration_sec:
            annot["time_max_offset"] = sound.file_duration_sec
            print('The annotation end time was adjusted as it exceeded the end time of the audio recording.')

        if annot["time_min_offset"] < sound.file_duration_sec: # if annotation is completely outside teh recording -> discard

            # define duration of noise window
            if self.noise_win_sec == "auto":
                half_noise_win_dur = annot.duration / 2
            else:
                half_noise_win_dur = self.noise_win_sec / 2

            # define left noise window
            noise_left_start = annot["time_min_offset"] - half_noise_win_dur
            if noise_left_start < 0:
                noise_left_end = annot["time_min_offset"]
                noise_left_start = 0
            else:
                noise_left_end = annot["time_min_offset"]

            # define right noise window
            noise_right_start = annot["time_max_offset"]
            noise_right_end = noise_right_start + half_noise_win_dur
            if noise_right_end > sound.file_duration_sec:
                noise_right_end = sound.file_duration_sec

            # load sound data chunk
            try:
                sound.read(chunk=[noise_left_start, noise_right_end], unit="sec")
            except:
                print(annot)
                raise Exception("error with time boundaries")

            # bandpass filter
            try:
                sound.filter(
                    "bandpass",
                    cutoff_frequencies=[
                        annot["frequency_min"],
                        annot["frequency_max"],
                    ],
                    order=10,
                    verbose=False,
                )
            except:
                print(annot)
                raise Exception("error with frequency filtering")
            sound.normalize()

            # calculate energies
            times_samp = np.round(
                np.dot(
                    [
                        noise_left_start,
                        noise_left_end,
                        noise_right_start,
                        noise_right_end,
                    ],
                    sound.waveform_sampling_frequency,
                )
            )
            times_samp = times_samp - times_samp[0]
            noise_left = sound.waveform[int(times_samp[0]) : int(times_samp[1])]
            sig = sound.waveform[int(times_samp[1]) : int(times_samp[2])]
            noise_right = sound.waveform[int(times_samp[2]) : int(times_samp[3])]
            # noise_pw = (sum(noise_left**2) + sum(noise_right**2)) / (
            #     len(noise_left) + len(noise_right)
            # )
            # sig_pw = sum(sig**2) / len(sig)

            noise_rms = np.sqrt(
                (sum(noise_left**2) + sum(noise_right**2))
                / (len(noise_left) + len(noise_right))
            )
            sig_rms = np.sqrt(sum(sig**2) / len(sig))

            # noise_var = np.sqrt((sum(noise_left**2) + sum(noise_right**2)))

            # try:
            snr = 20 * np.log10(sig_rms / noise_rms)
            # snr = 10 * np.log10(sig_pw / noise_pw)
            # except:
            #    print("Error")
            #    snr = np.nan

            if debug:
                sound.plot(newfig=True, title=str(round(snr, 1)))
        else:
            print('Annotation outside of audio recording.')
            snr = np.nan
        # stack all features
        tmp = pd.DataFrame(
            {
                "uuid": [annot["uuid"]],
                "snr": [snr],
            }
        )
        return tmp




from ecosound.core.annotation import Annotation
from ecosound.core.tools import filename_to_datetime
#root_dir = r'C:\Users\xavier.mouy\Desktop\raick'
#annot = Annotation()
#annot.from_raven(root_dir)

# print('d')
#
# file =r'C:\Users\xavie\Desktop\WHOI_fish\data_training_and_evaluation\20220711T174119.wav'
# file2 = r"C:\Users\xavie\Desktop\WHOI_fish\data_training_and_evaluation_1bigfile\20220711T170024.wav"
#
# dep_file =r'C:\Users\xavier.mouy\Desktop\raick\deployment_info_test2.csv'
# date = filename_to_datetime(file)
#
#
# from ecosound.core.metadata import DeploymentInfo
# dep_info = DeploymentInfo()
# print()



annot_file = r'C:\Users\xavier.mouy\Desktop\raick\New folder'
annot = Annotation()
annot.from_raven(annot_file, class_header='Annotation')
print('s')



# -*- coding: utf-8 -*-
"""
Created on Wed May  5 16:17:09 2021

@author: xavier.mouy
"""
import os
import sys
sys.path.append(r'C:\Users\xavier.mouy\Documents\GitHub\ecosound') # Adds higher directory to python modules path.

from ecosound.core.tools import derivative_1d, find_peaks, envelop
from numpy import array, sign, zeros
from scipy.interpolate import interp1d
from matplotlib.pyplot import plot,show,grid

s = array([1,4,3,5,3,2,4,3,4,5,4,3,2,5,6,7,8,7,8]) #This is your noisy vector of values.

q_u = zeros(s.shape)
q_l = zeros(s.shape)

# #Prepend the first value of (s) to the interpolating values. This forces the model to use the same starting point for both the upper and lower envelope models.

# u_x = [0,]
# u_y = [s[0],]

# l_x = [0,]
# l_y = [s[0],]

# #Detect peaks and troughs and mark their location in u_x,u_y,l_x,l_y respectively.

# u_x, u_y = find_peaks(s,troughs=True)
# l_x, l_y = find_peaks(s,troughs=False)

# #Append the last value of (s) to the interpolating values. This forces the model to use the same ending point for both the upper and lower envelope models.
# u_x.append(len(s)-1)
# u_y.append(s[-1])
# l_x.append(len(s)-1)
# l_y.append(s[-1])

# #Fit suitable models to the data. Here I am using cubic splines, similarly to the MATLAB example given in the question.
# u_p = interp1d(u_x,u_y, kind = 'cubic',bounds_error = False, fill_value=0.0)
# l_p = interp1d(l_x,l_y,kind = 'cubic',bounds_error = False, fill_value=0.0)

# #Evaluate each model over the domain of (s)
# for k in range(0,len(s)):
#     q_u[k] = u_p(k)
#     q_l[k] = l_p(k)

q_u, q_l = envelop(s, interp='cubic')

#Plot everything
plot(s)
plot(q_u,'r')
plot(q_l,'g')
grid(True);show()


from ecosound.core.annotation import Annotation
from ecosound.core.tools import list_files, filename_to_datetime
import datetime
import os

in_dir = r'C:\Users\xavier.mouy\Desktop\fichiers_nc'

files = list_files(in_dir,'.nc')
for file in files:
    annot = Annotation()
    annot.from_netcdf(file)
    for n in range(0,len(annot)):
        annot.data['time_min_date'].iloc[n] = filename_to_datetime(file)[0] + datetime.timedelta(seconds=annot.data.iloc[n]['time_min_offset'])
        annot.data['time_max_date'].iloc[n] = filename_to_datetime(file)[0] + datetime.timedelta(seconds=annot.data.iloc[n]['time_max_offset'])
    annot.to_netcdf(file)

print('done')





# -*- coding: utf-8 -*-
"""
Created on Tue Nov  8 10:55:51 2022

@author: xavier.mouy
"""

from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from ecosound.measurements.measurer_builder import MeasurerFactory
from ecosound.core.audiotools import Sound
import os
import numpy as np

annot_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20201102\Annotations_dataset_UK-SAMS-WestScotland-202011-N1 annotations.nc"
out_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20201102\Annotations_dataset_UK-SAMS-WestScotland-202011-N1 annotations_withSNR.nc"
#spectro_dir = r'D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20201102\spectrograms'

#annot_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20210522\Annotations_dataset_UK-SAMS-WestScotland-202105-N1 annotations.nc"
#out_file = r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\continuous_datasets\UK-SAMS-N1-20210522\Annotations_dataset_UK-SAMS-WestScotland-202105-N1 annotations_withSNR.nc"

noise_win_sec = 20 #"auto"  # 0.25

# load annotations
dataset = Annotation()
dataset.from_netcdf(annot_file)

# dataset.filter('label_class=="FS"', inplace=True)
# dataset.data = dataset.data.iloc[:100]
# dataset.update_audio_dir(r"D:\NOAA\2022_BC_fish_detector\manual_annotations")
# dataset.data.drop([0,1,2],inplace=True)

#dataset.data = dataset.data.iloc[2493:2495]
#dataset.data = dataset.data.to_frame()
#dataset.data.reset_index(drop=True, inplace=True)

# Meausrement
snr_measurer = MeasurerFactory("SNR", noise_win_sec=noise_win_sec)
measurements_snr = snr_measurer.compute(dataset, verbose=True, debug=False)

# measurements_snr.export_spectrograms(spectro_dir,
#                                      sanpling_rate_hz=2000,
#                                      file_prefix_field='snr',
#                                      )

measurements_snr.to_netcdf(out_file)

print("done")



# -*- coding: utf-8 -*-.
"""
Created on Tue Jan 21 13:04:58 2020

@author: xavier.mouy
"""
from ecosound.core.annotation import Annotation
import ecosound.core.tools
import pandas as pd
import xarray as xr
import os


class Measurement(Annotation):
    def __init__(
        self,
        measurer_name=None,
        measurer_version=None,
        measurements_name=None,
        measurements_parameters=None,
    ):
        """Measurement object.

        Object to "store" sound measurements. Inheritate all methods from the
        ecosound Annotaion class.

        Parameters
        ----------
        measurer_name : str, optional
            Name of the measurer that was used to calculate the measurements.
            The default is None.
        measurer_version : str, optional
            Version of the measurer that was used to calculate the measurements.
            The default is None.
        measurements_name : list of str, optional
            List with the name of each measurement. The default is None.
        measurements_parameters: dict, optional
            dict with lists of measurement parameters
        Returns
        -------
        None. ecosound Measurement object with a .data and .metadata dataframes

        """
        super(Measurement, self).__init__()
        metadata = {
            "measurer_name": measurer_name,
            "measurer_version": measurer_version,
            "measurements_name": [measurements_name],
            "measurements_parameters": [measurements_parameters],
        }
        self._metadata = pd.DataFrame(metadata)
        self.data = pd.concat(
            [self.data, pd.DataFrame(columns=metadata["measurements_name"][0])]
        )

    @property
    def metadata(self):
        """
        Return the metadata attribute.

        Includes adictionary with the measurer_name, measurer_version, and
        measurements_name.
        """
        return self._metadata

    def to_netcdf(self, file):
        """
        Write measurement data to a netcdf file.

        Write measurementss as .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        file : str
            Path of the netcdf file (.nc) to be written.

        Returns
        -------
        None.
        """

        if file.endswith(".nc") == False:
            file = file + ".nc"
        self._enforce_dtypes()
        meas = self.data
        meas.set_index("time_min_date", drop=False, inplace=True)
        meas.index.name = "date"
        dxr1 = meas.to_xarray()
        dxr1.attrs["datatype"] = "Measurement"
        dxr1.attrs[
            "measurements_name"
        ] = self.metadata.measurements_name.values[0]
        dxr1.attrs["measurer_name"] = self.metadata.measurer_name.values[0]
        dxr1.attrs["measurer_version"] = self.metadata.measurer_version.values[
            0
        ]
        try:
            dxr1.attrs["measurements_parameters"] = str(
                self.metadata.measurements_parameters[0]
            )
        except:
            pass
        dxr1.to_netcdf(file, engine="netcdf4", format="NETCDF4")

    def to_raven(
        self, outdir, outfile="Raven.Table.1.selections.txt", single_file=False
    ):
        """
        Write data to 1 or several Raven files.

        Write measurements as .txt files readable by the software Raven. Output
        files can be written in a single txt file or in several txt files (one
        per audio recording). In the latter case, output file names are
        automatically generated based on the audio file's name.

        Parameters
        ----------
        outdir : str
            Path of the output directory where the Raven files are written.
        outfile : str
            Name of the output file. Only used is single_file is True. The
            default is 'Raven.Table.1.selections.txt'.
        single_file : bool, optional
            If set to True, writes a single output file with all annotations.
            The default is False.

        Returns
        -------
        None.

        """
        cols = [
            "Selection",
            "View",
            "Channel",
            "Begin Time (s)",
            "End Time (s)",
            "Delta Time (s)",
            "Low Freq (Hz)",
            "High Freq (Hz)",
            "Begin Path",
            "File Offset (s)",
            "Begin File",
            "Class",
            "Sound type",
            "Software",
            "Confidence",
        ]
        
        # add measurements names
        cols = cols + self.metadata['measurements_name'][0]
        
        if len(self) > 0:
            if single_file:
                annots = [self.data]
            else:
                annots = [
                    pd.DataFrame(y)
                    for x, y in self.data.groupby(
                        "audio_file_name", as_index=False
                    )
                ]
            for annot in annots:
                annot.reset_index(inplace=True, drop=True)
                
                outdf = pd.DataFrame(columns =cols)
                # outdf = pd.DataFrame(
                #     {
                #         "Selection": 0,
                #         "View": 0,
                #         "Channel": 0,
                #         "Begin Time (s)": 0,
                #         "End Time (s)": 0,
                #         "Delta Time (s)": 0,
                #         "Low Freq (Hz)": 0,
                #         "High Freq (Hz)": 0,
                #         "Begin Path": 0,
                #         "File Offset (s)": 0,
                #         "Begin File": 0,
                #         "Class": 0,
                #         "Sound type": 0,
                #         "Software": 0,
                #         "Confidence": 0,
                #     },
                #     index=list(range(annot.shape[0])),
                # )
                
                
                outdf["Selection"] = range(1, annot.shape[0] + 1)
                outdf["View"] = "Spectrogram 1"
                outdf["Channel"] = annot["audio_channel"]
                outdf["Begin Time (s)"] = annot["time_min_offset"]
                outdf["End Time (s)"] = annot["time_max_offset"]
                outdf["Delta Time (s)"] = annot["duration"]
                outdf["Low Freq (Hz)"] = annot["frequency_min"]
                outdf["High Freq (Hz)"] = annot["frequency_max"]
                outdf["File Offset (s)"] = annot["time_min_offset"]
                outdf["Class"] = annot["label_class"]
                outdf["Sound type"] = annot["label_subclass"]
                outdf["Software"] = annot["software_name"]
                outdf["Confidence"] = annot["confidence"]
                outdf["Begin Path"] = [
                    os.path.join(x, y) + z
                    for x, y, z in zip(
                        annot["audio_file_dir"],
                        annot["audio_file_name"],
                        annot["audio_file_extension"],
                    )
                ]
                outdf["Begin File"] = [
                    x + y
                    for x, y in zip(
                        annot["audio_file_name"], annot["audio_file_extension"]
                    )
                ]
                outdf = outdf.fillna(0)
                # add neasurements
                outdf[self.metadata['measurements_name'][0]]=annot[self.metadata['measurements_name'][0]]                
                if single_file:
                    outfilename = os.path.join(outdir, outfile)
                else:
                    outfilename = os.path.join(
                        outdir,
                        str(annot["audio_file_name"].iloc[0])
                        + str(annot["audio_file_extension"].iloc[0])
                        + ".chan"
                        + str(annot["audio_channel"].iloc[0])
                        + ".Table.1.selections.txt",
                    )                                                
                outdf.to_csv(
                    outfilename,
                    sep="\t",
                    encoding="utf-8",
                    header=True,
                    columns=cols,
                    index=False,
                )
        else:
            # No annotation => write file with header only
            outfilename = os.path.join(outdir, outfile)
            header = "\t".join(cols)
            f = open(outfilename, "w")
            f.write(header)
            f.close()
    

    def from_netcdf(self, file, verbose=False):
        """
        Import measurement data from a netcdf file.

        Load measurements from a .nc file. This format works well with xarray
        and Dask. Only "Measurement" netcdf files created with to_netcdf can
        be imported. all other netcdf files will return an error.

        Parameters
        ----------
        file : str
            Path of the nc file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple
            files. If 'files' is a folder, all files in that folder ending with
            '.nc' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(file) is str:
            if os.path.isdir(file):
                file = ecosound.core.tools.list_files(
                    file,
                    ".nc",
                    recursive=False,
                    case_sensitive=True,
                )
                if verbose:
                    print(len(file), "files found.")
            else:
                file = [file]
        self.data, self._metadata = self._import_netcdf_files(file)
        self.check_integrity(verbose=verbose)

    def _import_netcdf_files(self, files):
        """Import one or several netcdf files to a Panda datafrane."""
        assert type(files) in (
            str,
            list,
        ), "Input must be of type str (single \
            file or directory) or list (multiple files)"
        # Import all files to a dataframe
        tmp = []
        for idx, file in enumerate(files):
            dxr = xr.open_dataset(file)
            if dxr.attrs["datatype"] == "Measurement":
                if idx == 0:
                    measurer_name = dxr.measurer_name
                    measurer_version = dxr.measurer_version
                    measurements_name = dxr.measurements_name
                    try:
                        measurements_parameters = eval(
                            dxr.measurements_parameters
                        )
                    except:
                        measurements_parameters = None
                ## check measurere name and version
                if (dxr.measurer_name == measurer_name) & (
                    dxr.measurer_version == measurer_version
                ):
                    tmp2 = dxr.to_dataframe()
                    tmp2.reset_index(inplace=True)
                else:
                    raise ValueError(
                        file
                        + "Not all files were not generated from the same measurer type and version."
                    )
            else:
                raise ValueError(file + "Not a Measurement file.")
            tmp.append(tmp2)
        data = pd.concat(tmp, ignore_index=True, sort=False)
        data.reset_index(inplace=True, drop=True)
        metadata = {
            "measurer_name": measurer_name,
            "measurer_version": measurer_version,
            "measurements_name": [measurements_name],
            "measurements_parameters": [measurements_parameters],
        }
        metadata = pd.DataFrame(metadata)
        return data, metadata

    def __add__(self, other):
        """Concatenate data from several Measurement objects."""
        assert (
            type(other) is ecosound.core.measurement.Measurement
        ), "Object type not\
            supported. Can only concatenate Measurement objects together."
        assert (
            other.metadata["measurer_name"].values[0]
            == self.metadata["measurer_name"].values[0]
        ), "Can't concatenate measurements made from different measurers."
        assert (
            other.metadata["measurer_version"].values[0]
            == self.metadata["measurer_version"].values[0]
        ), "Can't concatenate measurements made from different versions of measurers."
        self._enforce_dtypes()
        other._enforce_dtypes()
        self.data = pd.concat(
            [self.data, other.data], ignore_index=True, sort=False
        )
        return self



# -*- coding: utf-8 -*-
"""
Created on Fri Feb 21 16:43:47 2020

@author: xavier.mouy
"""

from .grapher_builder import BaseClass

try:
    from ecosound.core.annotation import Annotation
except ImportError:
    pass


import numpy as np
import pandas as pd
import ecosound
import matplotlib.pyplot as plt
import matplotlib.colors as mc  # For the legend
import matplotlib.dates as mdates
from mpl_toolkits.axes_grid1 import make_axes_locatable
import datetime as dt

# import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas

# from matplotlib.figure import Figure
import matplotlib

# import numpy as np


class AnnotHeatmap(BaseClass):
    """A Grapher class to visualize annotation data.

    Display heatmap of annotations (date vs time-of-day).

    The HeatmapPlotter grapher must be instantiated using the GrapherFactory
    with the positional argument 'HeatmapPlotter':

    from ecosound.visualization.grapher_builder import GrapherFactory
    graph = GrapherFactory('HeatmapPlotter',kwargs)

    Attributes
    ----------
    name : str
        Name of the grapher
    version : str
        Version of the grapher
    fig_size : (float, float)
        Width, height of the figure, in inches.
    title : [str, str, ...] or str
        Title(s) of the figure/subplots.
    colormap : str
        Color map palette spectrogram. Uses names from Matplotlib. The default
        is 'viridis'
    colorbar_label: str
        Label to use for the clolorbar.
    integration_time: str
        Integration time for the time aggregates.
    date_format: str
        Date format to use for the x-axis tick labels.
    is_binary: bool
        Makes binary aggreagates rather than counts.
    norm_value: int
        Maximum value to use for the colomap.
    share_xaxis: bool
        Share xaxis in case of subplots (not implememnted yet)

    Methods
    -------
    add_data(Annotation obj)
        Add Annotation object to plot.
    show(display=True)
        Display graph.
    to_file(filename)
        Save graph to file.
    """

    grapher_parameters = (
        "date_format",
        "colorbar_label" "integration_time",
        "is_binary",
        "norm_value",
        "fig_size",
        "share_xaxis",
        "title",
        "colormap",
    )

    def __init__(self, *args, **kwargs):
        """
        Initialize grapher object.

        Initialize the grapher object to display sound data.

        Parameters
        ----------
        *args : str
            Do not use. Only used by the GrapherFactory.
        integration_time: str, optional
            Integration time for the aggregate. Uses the pandas offset aliases
            (i.e. '2H'-> 2 hours, '15min'=> 15 minutes, '1D'-> 1 day) see pandas
            documnentation here:https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
        fig_size : (float, float), optional
            Width, height of the figure, in inches. The default is (16, 4).
        share_xaxis : bool, optional
            Share x axis for all subplots. The default is False.
        is_binary: bool, optional
            If set to True, calculates the aggregates in term on presence (1)
            or absence (0). The default is False.
        title : [str, str, ...], str, None,  optional
            Title(s) of the figure/subplots. If a str, a single title is set
            for the entire figure at the top. If a list of str, each element i
            of the list defines the title for the subplot of index i. The
            number of elements in the list can equal or smaller than the number
            of subplots in the graph. A number of elements in teh list greater
            than the number of subplot will return an error. If set to None, no
            title will be displayed. The default is None.
        colormap : str, optional
            Color map for the spectrogram. Uses names from Matplotlib (see
            https://matplotlib.org/stable/tutorials/colors/colormaps.html). The
            default is 'viridis'
        colorbar_label: str, optional
            Label for the color bar. If set to 'auto', the colorbar label is
            automatically defined as "Annotations" or "Detections" based on
            the field "is_detector" of the Annotation object, and "(count)" or
            "(presence/absence)" based on the grapher attribute 'is_binary'.
            The default is 'auto'.
        norm_value: flaat, None
            Maximum value to use for the colobar. The default is None.
        date_format: str, optional
            Date format to use for the x-axis tick labels. Uses standard python
            datetime format codes (see https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes)
            The default is '%d-%b-%Y'.

        Returns
        -------
        None.

        """
        # Initialize all grapher parameters to None
        self.__dict__.update(
            dict(
                zip(
                    self.grapher_parameters,
                    [None] * len(self.grapher_parameters),
                )
            )
        )
        # Define default values:
        self.date_format = "%d-%b-%Y"
        self.colorbar_label = "auto"
        self.share_xaxis = False
        self.integration_time = "1H"
        self.is_binary = False
        self.norm_value = None
        self.fig_size = (16, 4)
        self.title = None
        self.colormap = "viridis"
        self.data = []
        # Unpack kwargs as grapher parameters if provided on instantiation
        self.__dict__.update(**kwargs)

    @property
    def name(self):
        """Return name of the grapher."""
        grapher_name = "AnnotHeatmap"
        return grapher_name

    @property
    def version(self):
        """Return version of the grapher."""
        version = "0.1"
        return version

    def add_data(self, *args):
        """
        Define annotation data to plot.

        Add Annotation objects to plot. There is no restriction on the number
        of object to add. Each object will be displayed on a different subplot
        of the same figure. The order in which the objects are passed to this
        method defines the order in which they are displayed (i.e. subplots).
        When this method is called several times; objects are appended to the
        existing list of objects from previous calls.

        Parameters
        ----------
        *args : Annotation objects
            Annotation objects to display.

        Raises
        ------
        ValueError
            If no input arguments are provided.
            If the input arguments are not Annotation objects.

        Returns
        -------
        None. Updated grapher object.

        """
        if len(args) < 1:
            raise ValueError("There must be at least one input argument")
        # Check  type of each input arguments
        self._stack_data(args)

    def show(self, display=True):
        """
        Display graph on screen.

        Display graph made of vertical subplots for each of the Annotation
        objects defined by the add_data method. The order in which the objects
        have been defined by add_data defines the order in which subplot are
        displayed.

        Parameters
        ----------
        display : bool, optional
            Display figure on screen. The default is True.

        Raises
        ------
        ValueError
            If no data have been defined (see add_data method).

        Returns
        -------
        fig : Figure instance
            Figure object.
        ax : Axis instance
            Axis object with all subplots.

        """
        if len(self.data) == 0:
            raise ValueError(
                "No data to plot. Use method .add_data to define"
                " the data to plot"
            )
        # Display plot on screen?
        if display:
            matplotlib.use("Qt5Agg")
        else:
            matplotlib.use("Agg")
        # Define new figure and subplots
        nb_plots = len(self.data)
        fig, ax = plt.subplots(
            nb_plots,
            1,
            figsize=self.fig_size,
            sharex=self.share_xaxis,
            constrained_layout=False,
        )  # gridspec_kw={'hspace': self.hspace}
        # Subplot titles
        titles = [None] * nb_plots
        if self.title is None:  # no titles
            pass
        if type(self.title) is str:
            titles[0] = self.title
        if type(self.title) is list:
            if len(self.title) > nb_plots:
                raise ValueError("More titles than subplots")
            else:
                titles[0 : len(self.title) - 1] = self.title
        # normalization values
        norm_values = [None] * nb_plots
        if self.norm_value is None:  # no titles
            pass
        if (type(self.norm_value) is float) or (type(self.norm_value) is int):
            norm_values = [self.norm_value] * nb_plots
        if type(self.norm_value) is list:
            if len(self.norm_value) > nb_plots:
                raise ValueError("More norm_value than subplots")
            else:
                norm_values[0 : len(self.norm_value) - 1] = self.norm_value
        # Plot data
        for idx, data in enumerate(self.data):
            if nb_plots == 1:
                current_ax = ax
            else:
                current_ax = ax[idx]
            if data["type"] == "annotation":
                self._plot_heatmap(
                    data["data"],
                    current_ax,
                    title=titles[idx],
                    norm_value=norm_values[idx],
                )
            # only dipslay x label of bottom plot if shared axes
            if self.share_xaxis and (idx != nb_plots - 1):
                current_ax.set_xlabel("")
        # fig.tight_layout()
        # fig.set_tight_layout(True)
        return fig, ax

    def to_file(self, filename):
        """
        Save the figure to an image file.

        Parameters
        ----------
        filename : str
            Full path of the image file to write.

        Returns
        -------
        None.

        """
        fig, _ = self.show(display=False)
        fig.savefig(
            filename,
            transparent=False,
            bbox_inches="tight",
        )

    def _stack_data(self, args):
        """Stack data to be plotted."""
        for idx, arg in enumerate(args):
            if isinstance(arg, ecosound.core.annotation.Annotation):
                self.data.append({"data": arg, "type": "annotation"})
            else:
                raise ValueError(
                    "Type of input argument not recognized."
                    "Accepted object types: Annotation"
                )

    def _plot_heatmap(self, annot, current_ax, title=None, norm_value=None):
        """Plot heatmap on the current axis"""
        # calulate 1D aggreagate
        data_grid = annot.calc_time_aggregate_2D(
            integration_time=self.integration_time, is_binary=self.is_binary
        )
        axis_date = data_grid.columns.to_list()
        # Plot matrix
        x_lims = mdates.date2num(
            [axis_date[0], axis_date[-1] + dt.timedelta(1)]
        )
        y_lims = mdates.date2num(
            [
                dt.datetime.combine(axis_date[0], dt.time(0, 0)),
                dt.datetime.combine(axis_date[0], dt.time(23, 59, 59)),
            ]
        )
        if self.is_binary:
            norm_value = 1
        im = current_ax.imshow(
            data_grid,
            extent=[x_lims[0], x_lims[1], y_lims[0], y_lims[1]],
            vmin=0,
            vmax=norm_value,
            aspect="auto",
            origin="lower",
            cmap=self.colormap,
        )
        current_ax.xaxis_date()
        current_ax.yaxis_date()
        current_ax.xaxis.set_major_formatter(
            mdates.DateFormatter(self.date_format)
        )
        current_ax.yaxis.set_major_formatter(mdates.DateFormatter("%H:%M"))
        divider = make_axes_locatable(current_ax)
        cax = divider.append_axes("right", size="2%", pad=0.05)
        cbar = plt.colorbar(im, cax=cax)
        current_ax.set_title(title)
        current_ax.set_ylabel("Time of day")
        if self.colorbar_label == "auto":
            if bool(annot.data.from_detector.iloc[0]):
                if self.is_binary:
                    cbar.set_label("Detections \n(presence)")
                else:
                    cbar.set_label("Detections \n(count)")
            else:
                if self.is_binary:
                    cbar.set_label("Annotations \n(presence)")
                else:
                    cbar.set_label("Annotations \n(count)")
        else:
            cbar.set_label(self.colorbar_label)
        return



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 13:35:34 2022

@author: xavier.mouy
"""

import sys
sys.path.append(r"C:\Users\xavier.mouy\Documents\GitHub\ecosound") # Adds higher directory to python modules path.
from ecosound.core.annotation import Annotation

infile = r'C:\Users\xavier.mouy\Documents\GitHub\ecosound\data\sqlite_annotations/read/detections1.sqlite'

# load sqlite tables
annot = Annotation()
annot.from_sqlite(infile, verbose=True)

# calculate aggregates
agg_1D = annot.calc_time_aggregate_1D(integration_time='1H',is_binary=True) 

agg_2D = annot.calc_time_aggregate_2D(integration_time='15min',is_binary=False)



# plot
annot.heatmap(norm_value = 30,integration_time='1H',is_binary=False)
#plt.savefig(os.path.join(annot_dir, 'heatmap_figure.png'))



# -*- coding: utf-8 -*-
"""
Created on Mon Mar 22 16:24:27 2021

@author: xavier.mouy
"""
import sys
sys.path.append(r"C:\Users\xavier.mouy\Documents\GitHub\ecosound")  # Adds higher directory to python modules path.
from ecosound.core.measurement import Measurement

infile = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Herring_DFO\New folder\5042.200306203002.wav.nc'
outdir = r'C:\Users\xavier.mouy\Documents\PhD\Projects\Herring_DFO\New folder'

mes = Measurement()
mes.from_netcdf(infile)

mes.to_pamlab(outdir)
mes.to_raven(outdir)


# -*- coding: utf-8 -*-
"""
Created on Tue Jun 28 09:38:49 2022

@author: xavier.mouy
"""

from ecosound.core.annotation import Annotation
from ecosound.core.tools import list_files, filename_to_datetime
import ecosound
import matplotlib.pyplot as plt
import os
import sqlite3
import pandas as pd
## ############################################################################
## Input parameters ###########################################################

detec_dir = r'\\nefscfile\PassiveAcoustics\Stellwagen_Old\DETECTORS\Pile_driving\Pile_driving_NEFSC_MA-RI_202304_NS02_ST'
audio_dir=r'\\nefscfile\PassiveAcoustics\Stellwagen_Old\ACOUSTIC_DATA\BOTTOM_MOUNTED\NEFSC_MA-RI\NEFSC_MA-RI_202304_NS02\NEFSC_MA-RI_202304_NS02_ST\6075_64kHz_UTC'
deployment_name = 'NEFSC_MA-RI_202304_NS02_ST'
audio_file_ext = '.wav'
aggregate_time_offset = 0
## ############################################################################
## ############################################################################

out_dir=os.path.join(detec_dir,'detections_summary')

#creates output dir
if os.path.exists(out_dir) == False:
    os.mkdir(out_dir)

# import detection results
print('Loading detections')
annot = Annotation()
annot.from_raven(detec_dir,recursive=True)

### update path
annot.data.loc[:,'audio_sampling_frequency']=0
annot.data.loc[:,'audio_bit_depth']=0
annot.data.loc[:,'audio_file_dir']=audio_dir

### save updated annot as netcdf file
annot.to_netcdf(os.path.join(detec_dir,'detections.nc'))

# # list files
# print('Creating detection summary spreadsheet')
audio_files = list_files(audio_dir,suffix=audio_file_ext,recursive=True)
dates = filename_to_datetime(audio_files)
# files_list=[]
# detec_count=[]
# for audio_file in audio_files:
#     name = os.path.basename(os.path.splitext(audio_file)[0])
#     files_list.append(name)
#     temp = annot.filter('audio_file_name == "' + name + '"')
#     detec_count.append(len(temp))
# table = pd.DataFrame({'file':files_list, 'detections': detec_count})
# table.to_csv(os.path.join(out_dir,'merged_detections_per_files.csv'),index=False)


# Create hourly agggregate.
print("Creating detections hourly aggregates...")
hourly_counts = annot.calc_time_aggregate_1D(
    integration_time="H",
    resampler="count",
    start_date= str(min(dates)),
    end_date= str(max(dates)),
)

hourly_counts.rename(columns={"value": "Detections"}, inplace=True)
hourly_counts.to_csv(
    os.path.join(out_dir, deployment_name + "_hourly_counts.csv"),
    index_label="Date (UTC" + str(aggregate_time_offset) + ")",
)


# create spectrograms
print('Creating spectrograms')
annot.export_spectrograms(
        out_dir,
        time_buffer_sec=4,
        spectro_unit="sec",
        spetro_nfft=0.1,
        spetro_frame=0.1,
        spetro_inc=0.03,
        freq_min_hz=10,
        freq_max_hz=1000,
        sanpling_rate_hz=4000,
        filter_order=8,
        filter_type="iir",
        fig_size=(20, 10),
        deployment_subfolders=False,
        date_subfolders=True,
        date_subfolders_format = "%Y-%m-%d_%H",
        #file_subfolder=True,
        #file_name_field="time_max_date",
        file_name_field="time_min_offset",
        file_prefix_field="audio_file_name",
        channel=None,
        colormap="viridis",#"Greys",  # "viridis",
        save_wav=False,
    )






import os
import pandas as pd
from ecosound.core.annotation import Annotation
from ecosound.core.audiotools import Sound
from ecosound.core.tools import list_files

data_dir = r'\\stellwagen.nefsc.noaa.gov\stellwagen\ACOUSTIC_DATA\BOTTOM_MOUNTED\NEFSC_GOM\NEFSC_GOM_202205_USTR01\6550_48kHz_UTC'
detec_dir = r'\\stellwagen.nefsc.noaa.gov\stellwagen\STAFF\Lindsey\humpback_detector_outputs\NEFSC_GOM_202205_USTR01_2023-02-24_16-08-13\thresh_0.9\NEFSC_GOM_202205_USTR01\6550_48kHz_UTC'
outfilename =r'\\stellwagen.nefsc.noaa.gov\stellwagen\STAFF\Lindsey\humpback_detector_outputs\NEFSC_GOM_202205_USTR01_2023-02-24_16-08-13\thresh_0.9\NEFSC_GOM_202205_USTR01\6550_48kHz_UTC\merged_detections\merged_tables.txt'
file_ext = '.wav'



# define dates and time offsets
data_files_dates = list_files(data_dir,suffix=file_ext)
data_files_dates.sort()
filenames=[]
files_dur =[]
for idx, file in enumerate(data_files_dates):
    s = Sound(file);
    filenames.append(os.path.split(file)[1][0:-len(file_ext)])
    files_dur.append(s.file_duration_sec)
files_dur = [0] + files_dur[0:-1]
filename_nchar = len(filenames[0])

file_offset_sec=[]
for idx, dur in enumerate(files_dur):
    file_offset_sec.append(sum(files_dur[0:idx+1]))

# load all text files
raven_files = list_files(detec_dir,suffix='.txt')
for idx, file in enumerate(raven_files):
    tmp = pd.read_csv(file, delimiter="\t")
    file_idx = filenames.index(os.path.split(file)[1][0:filename_nchar])
    tmp['Begin Time (s)'] = tmp['Begin Time (s)'] + file_offset_sec[file_idx]
    tmp['End Time (s)'] = tmp['End Time (s)'] + file_offset_sec[file_idx]
    if idx == 0:
        data = tmp
    else:
        data = pd.concat([data, tmp], ignore_index=True, sort=False)

data['Selection'] = list(range(1,len(data)+1))
data.drop(columns=['Selection.1'],inplace=True)

data.to_csv(outfilename,
            sep="\t",
            encoding="utf-8",
            header=True,
            #columns=cols,
            index=False,
            )



# -*- coding: utf-8 -*-
"""
Created on Wed Mar  2 08:22:48 2022

@author: xavier.mouy
"""

# from ecosound.core.measurement import Measurement
from ecosound.core.annotation import Annotation

import datetime
import os

in_dir = (
    r"D:\NOAA\2022_Minke_whale_detector\manual_annotations\discrete_datasets"
)
file = "Annotations_dataset_NN-MW-HB-HK_20221213T142558.nc"
aggregate_time_offset = 0

# Load dataset
print("Loading detections...")
dataset = Annotation()
# dataset.from_sqlite(os.path.join(in_dir, file))
dataset.from_netcdf(os.path.join(in_dir, file))


# Filter
print("Filtering detections...")
dataset.filter("label_class=='MW'", inplace=True)
# dataset.filter("label_class!='HK'", inplace=True)
# dataset.filter("label_class=='NN'", inplace=True)
# dataset.filter(
#     "deployment_ID=='USA-NEFSC-MA-RI-202001-NS01'| deployment_ID=='USA-NEFSC-MA-RI-202202-NS02'",
#     inplace=True,
# )


# Create spectrograms and wav files
print("Extracting detection spectrograms...")
out_dir = os.path.join(in_dir, "extracted_detections")
if os.path.isdir(out_dir) == False:
    os.mkdir(out_dir)

dataset.export_spectrograms(
    out_dir,
    time_buffer_sec=5,
    spectro_unit="sec",
    spetro_nfft=0.256,
    spetro_frame=0.256,
    spetro_inc=0.03,
    freq_min_hz=0,
    freq_max_hz=1000,
    sanpling_rate_hz=2000,
    filter_order=8,
    filter_type="iir",
    fig_size=(15, 10),
    deployment_subfolders=True,
    date_subfolders=False,
    file_name_field="uuid",
    # file_name_field="audio_file_name",
    file_prefix_field="confidence",
    channel=0,
    colormap="Greys",  # "viridis",
    save_wav=True,
)

print("Done!")



# -*- coding: utf-8 -*-.
"""
Created on Tue Jan 21 13:04:58 2020

@author: xavier.mouy
"""

import pandas as pd
import xarray as xr
import numpy as np
import os
import uuid
import warnings
import ecosound.core.tools
import ecosound.core.decorators
import sqlite3
from ecosound.core.metadata import DeploymentInfo
from ecosound.visualization.grapher_builder import GrapherFactory
from ecosound.core.spectrogram import Spectrogram
from ecosound.core.audiotools import Sound
import copy
import csv
import datetime
import re
import warnings
from tqdm import tqdm


class Annotation:
    """
    A class used for manipulating annotation data.

    The Annotation object stores from both manual analysis annotations
    collected with software like PAMlab and Raven, and outputs from
    automated detectors and classifiers.

    Attributes
    ----------
    data : pandas DataFrame
        Annotation DataFranme.

    Methods
    -------
    check_integrity(verbose=False, time_duplicates_only=False)
        Check integrity of Annotation object.
    from_raven(files, class_header='Sound type',subclass_header=None,
               verbose=False)
        Import annotation data from 1 or several Raven files.
    to_raven(outdir, single_file=False)
        Write annotation data to one or several Raven files.
    from_pamlab(files, verbose=False)
        Import annotation data from 1 or several PAMlab files.
    to_pamlab(outdir, single_file=False)
        Write annotation data to one or several Raven files.
    from_parquet(file)
        Import annotation data from a Parquet file.
    to_parquet(file)
        Write annotation data to a Parquet file.
    from_netcdf(file)
        Import annotation data from a netCDF4 file.
    to_netcdf(file)
        Write annotation data to a netCDF4 file.
    insert_values(**kwargs)
        Manually insert values for given Annotation fields.
    insert_metadata(deployment_info_file)
        Insert metadata information to the annotation from a
        deployment_info_file.
    filter_overlap_with(annot, freq_ovp=True, dur_factor_max=None,
                        dur_factor_min=None,ovlp_ratio_min=None,
                        remove_duplicates=False,inherit_metadata=False,
                        filter_deploymentID=True, inplace=False)
        Filter annotations overalaping with another set of annotations.
    update_audio_dir(new_data_dir)
        Update path of audio files.
    get_labels_class()
        Return all unique class labels.
    get_labels_subclass()
        Return all unique subclass labels.
    get_fields()
        Return list with all annotations fields.
    summary(rows='deployment_ID',columns='label_class')
        Produce a summary pivot table with the number of annotations for two
        given annotation fields.
    __add__()
        Concatenate data from annotation objects uisng the + sign.
    __len__()
        Return number of annotations.
    """

    def __init__(self):
        """
        Initialize Annotation object.

        Sets all the annotation fields.:
            -'uuid': UUID,
                Unique identifier code
            -'from_detector': bool,
                True if data comes from an automatic process.
            -'software_name': str,
                Software name. Can be Raven or PAMlab for manual analysis.
            -'software_version': str,
                Version of the software used to create the annotations.
            -'operator_name': str,
                Name of the person responsible for the creation of the
                annotations.
            -'UTC_offset': float,
                Offset hours to UTC.
            -'entry_date': datetime,
                Date when the annotation was created.
            -'audio_channel': int,
                Channel number.
            -'audio_file_name': str,
                Name of the audio file.
            -'audio_file_dir': str,
                Directory where the audio file is.
            -'audio_file_extension': str,
                Extension of teh audio file.
            -'audio_file_start_date': datetime,
                Date of the audio file start time.
            -'audio_sampling_frequency': int,
                Sampling frequecy of the audio data.
            -'audio_bit_depth': int,
                Bit depth of the audio data.
            -'mooring_platform_name': str,
                Name of the moorig platform (e.g. 'glider','Base plate').
            -'recorder_type': str,
                Name of the recorder type (e.g., 'AMAR'), 'SoundTrap'.
            -'recorder_SN': str,
                Serial number of the recorder.
            -'hydrophone_model': str,
                Model of the hydrophone.
            -'hydrophone_SN': str,
                Serial number of the hydrophone.
            -'hydrophone_depth': float,
                Depth of the hydrophone in meters.
            -'location_name': str,
                Name of the deploymnet location.
            -'location_lat': float,
                latitude of the deployment location in decimal degrees.
            -'location_lon': float,
                longitude of the deployment location in decimal degrees.
            -'location_water_depth': float,
                Water depth at the deployment location in meters.
            -'deployment_ID': str,
                Unique ID of the deployment.
            -'frequency_min': float,
                Minimum frequency of the annotaion in Hz.
            -'frequency_max': float,
                Maximum frequency of the annotaion in Hz.
            -'time_min_offset': float,
                Start time of the annotaion, in seconds relative to the
                begining of the audio file.
            -'time_max_offset': float,
                Stop time of the annotaion, in seconds relative to the
                begining of the audio file.
            -'time_min_date': datetime,
                Date of the annotation start time.
            -'time_max_date': datetime,
                Date of the annotation stop time.
            -'duration': float,
                Duration of the annotation in seconds.
            -'label_class': str,
                label of the annotation class (e.g. 'fish').
            -'label_subclass': str,
                label of the annotation subclass (e.g. 'grunt')
            'confidence': float,
                Confidence of the classification.

        Returns
        -------
        Annotation object.

        """
        self.data = pd.DataFrame(
            {
                "uuid": [],
                "from_detector": [],  # True, False
                "software_name": [],
                "software_version": [],
                "operator_name": [],
                "UTC_offset": [],
                "entry_date": [],
                "audio_channel": [],
                "audio_file_name": [],
                "audio_file_dir": [],
                "audio_file_extension": [],
                "audio_file_start_date": [],
                "audio_sampling_frequency": [],
                "audio_bit_depth": [],
                "mooring_platform_name": [],
                "recorder_type": [],
                "recorder_SN": [],
                "hydrophone_model": [],
                "hydrophone_SN": [],
                "hydrophone_depth": [],
                "location_name": [],
                "location_lat": [],
                "location_lon": [],
                "location_water_depth": [],
                "deployment_ID": [],
                "frequency_min": [],
                "frequency_max": [],
                "time_min_offset": [],
                "time_max_offset": [],
                "time_min_date": [],
                "time_max_date": [],
                "duration": [],
                "label_class": [],
                "label_subclass": [],
                "confidence": [],
            }
        )
        self._enforce_dtypes()

    def check_integrity(
        self, verbose=False, ignore_frequency_duplicates=False
    ):
        """
        Check integrity of Annotation object.

        Tasks performed:
            1- Check that start time < stop time
            2- Check that min frequency < max frequency
            3- Remove duplicate entries based on time and frequency, filename,
               labels and filenames

        Parameters
        ----------
        verbose : bool, optional
            Print summary of the duplicate entries deleted.
            The default is False.
        ignore_frequency_duplicates : bool, optional
            If set to True, doesn't consider frequency values when deleting
            duplicates. It is useful when data are imported from Raven.
            The default is False.

        Raises
        ------
        ValueError
            If annotations have a start time > stop time
            If annotations have a min frequency > max frequency

        Returns
        -------
        None.

        """
        # Drop all duplicates
        count_start = len(self.data)
        if ignore_frequency_duplicates:  # doesn't use frequency boundaries
            self.data = self.data.drop_duplicates(
                subset=[
                    "time_min_offset",
                    "time_max_offset",
                    "label_class",
                    "label_subclass",
                    "audio_file_name",
                ],
                keep="first",
            ).reset_index(drop=True)
        else:  # remove annot with exact same time AND frequency boundaries
            self.data = self.data.drop_duplicates(
                subset=[
                    "time_min_offset",
                    "time_max_offset",
                    "frequency_min",
                    "frequency_max",
                    "label_class",
                    "label_subclass",
                    "audio_file_name",
                ],
                keep="first",
            ).reset_index(drop=True)
        count_stop = len(self.data)
        if verbose:
            print("Duplicate entries removed:", str(count_start - count_stop))
        # Check that start and stop times are coherent (i.e. t2 > t1)
        time_check = self.data.index[
            self.data["time_max_offset"] < self.data["time_min_offset"]
        ].tolist()
        if len(time_check) > 0:
            raise ValueError(
                "Incoherent annotation times (time_min > time_max). \
                 Problematic annotations:"
                + str(time_check)
            )
        # Check that min and max frequencies are coherent (i.e. fmin < fmax)
        freq_check = self.data.index[
            self.data["frequency_max"] < self.data["frequency_min"]
        ].tolist()
        if len(freq_check) > 0:
            raise ValueError(
                "Incoherent annotation frequencies (frequency_min > \
                frequency_max). Problematic annotations:"
                + str(freq_check)
            )

        # check that there are not uuid duplicates
        idx = self.data.duplicated(subset=["uuid"])
        dup_idxs = idx[idx == True].index
        for dup_idx in dup_idxs:
            self.data.loc[dup_idx, "uuid"] = str(uuid.uuid4())
        if len(dup_idxs) > 0:
            if verbose:
                print(
                    len(dup_idxs),
                    " UUID duplicates were found and regenerated.",
                )
        if verbose:
            print("Integrity test succesfull")

    def from_raven(
        self,
        files,
        class_header="Sound type",
        subclass_header=None,
        confidence_header=None,
        recursive=False,
        verbose=False,
    ):
        """
        Import data from 1 or several Raven files.

        Load annotation tables from .txt files generated by the software Raven.

        Parameters
        ----------
        files : str, list
            Path of the txt file(s) to import. Can be a str if importing a single
            file. Needs to be a list if importing multiple files. If 'files' is
            a folder, all files in that folder ending with '.selections.txt'
            will be imported.
        class_header : str, optional
            Name of the header in the Raven file corresponding to the class
            name. The default is 'Sound type'.
        subclass_header : str, optional
            Name of the header in the Raven file corresponding to the subclass
            name. The default is None.
        confidence_header : str, optional
            Name of the header in the Raven file corresponding to the confidence
            value. The default is None.
        recursive : bool, optional
            If set to True, goes rcursively through all subfolders. The default
            is False.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(files) is not list:
            if os.path.isdir(files):
                files = ecosound.core.tools.list_files(
                    files,
                    ".selections.txt",
                    recursive=recursive,
                    case_sensitive=True,
                )
                if verbose:
                    print(len(files), "annotation files found.")
        data = Annotation._import_csv_files(files)
        columns = data.columns.to_list()

        if "Begin Path" in columns:
            files_timestamp = ecosound.core.tools.filename_to_datetime(
                data["Begin Path"].tolist()
            )
            self.data["audio_file_name"] = data["Begin Path"].apply(
                lambda x: os.path.splitext(os.path.basename(x))[0]
            )
            self.data["audio_file_dir"] = data["Begin Path"].apply(
                lambda x: os.path.dirname(x)
            )
            self.data["audio_file_extension"] = data["Begin Path"].apply(
                lambda x: os.path.splitext(x)[1]
            )
        elif "Begin File" in columns:
            if verbose:
                print(
                    "'Begin Path' not found using 'Begin File' to retriev timestamps"
                )
            files_timestamp = ecosound.core.tools.filename_to_datetime(
                data["Begin File"].tolist()
            )
            self.data["audio_file_name"] = data["Begin File"].apply(
                lambda x: os.path.splitext(os.path.basename(x))[0]
            )
            self.data["audio_file_dir"] = None
            self.data["audio_file_extension"] = data["Begin File"].apply(
                lambda x: os.path.splitext(x)[1]
            )
        else:
            files_timestamp = None
            if verbose:
                print("Name of annotated audio files could not be found")
        self.data["audio_file_start_date"] = files_timestamp
        self.data["audio_channel"] = data["Channel"]
        self.data["time_min_offset"] = data["Begin Time (s)"]
        self.data["time_max_offset"] = data["End Time (s)"]
        self.data["time_min_date"] = pd.to_datetime(
            self.data["audio_file_start_date"]
            + pd.to_timedelta(self.data["time_min_offset"], unit="s")
        )
        self.data["time_max_date"] = pd.to_datetime(
            self.data["audio_file_start_date"]
            + pd.to_timedelta(self.data["time_max_offset"], unit="s")
        )
        self.data["frequency_min"] = data["Low Freq (Hz)"]
        self.data["frequency_max"] = data["High Freq (Hz)"]
        if class_header is not None:
            self.data["label_class"] = data[class_header]
        if subclass_header is not None:
            self.data["label_subclass"] = data[subclass_header]
        if confidence_header is not None:
            self.data["confidence"] = data[confidence_header]
        self.data["from_detector"] = False
        self.data["software_name"] = "raven"
        self.data["uuid"] = self.data.apply(
            lambda _: str(uuid.uuid4()), axis=1
        )
        self.data["duration"] = (
            self.data["time_max_offset"] - self.data["time_min_offset"]
        )
        self.check_integrity(verbose=verbose, ignore_frequency_duplicates=True)
        if verbose:
            print(len(self), "annotations imported.")

    def to_raven(
        self, outdir, outfile="Raven.Table.1.selections.txt", single_file=False
    ):
        """
        Write data to 1 or several Raven files.

        Write annotations as .txt files readable by the software Raven. Output
        files can be written in a single txt file or in several txt files (one
        per audio recording). In the latter case, output file names are
        automatically generated based on the audio file's name.

        Parameters
        ----------
        outdir : str
            Path of the output directory where the Raven files are written.
        outfile : str
            Name of the output file. Only used is single_file is True. The
            default is 'Raven.Table.1.selections.txt'.
        single_file : bool, optional
            If set to True, writes a single output file with all annotations.
            The default is False.

        Returns
        -------
        None.

        """
        cols = [
            "Selection",
            "View",
            "Channel",
            "Begin Time (s)",
            "End Time (s)",
            "Delta Time (s)",
            "Low Freq (Hz)",
            "High Freq (Hz)",
            "Begin Path",
            "File Offset (s)",
            "Begin File",
            "Class",
            "Sound type",
            "Software",
            "Confidence",
        ]
        if len(self) > 0:
            if single_file:
                annots = [self.data]
            else:
                annots = [
                    pd.DataFrame(y)
                    for x, y in self.data.groupby(
                        "audio_file_name", as_index=False
                    )
                ]
            for annot in annots:
                annot.reset_index(inplace=True, drop=True)
                outdf = pd.DataFrame(
                    {
                        "Selection": 0,
                        "View": 0,
                        "Channel": 0,
                        "Begin Time (s)": 0,
                        "End Time (s)": 0,
                        "Delta Time (s)": 0,
                        "Low Freq (Hz)": 0,
                        "High Freq (Hz)": 0,
                        "Begin Path": 0,
                        "File Offset (s)": 0,
                        "Begin File": 0,
                        "Class": 0,
                        "Sound type": 0,
                        "Software": 0,
                        "Confidence": 0,
                    },
                    index=list(range(annot.shape[0])),
                )
                outdf["Selection"] = range(1, annot.shape[0] + 1)
                outdf["View"] = "Spectrogram 1"
                outdf["Channel"] = annot["audio_channel"]
                outdf["Begin Time (s)"] = annot["time_min_offset"]
                outdf["End Time (s)"] = annot["time_max_offset"]
                outdf["Delta Time (s)"] = annot["duration"]
                outdf["Low Freq (Hz)"] = annot["frequency_min"]
                outdf["High Freq (Hz)"] = annot["frequency_max"]
                outdf["File Offset (s)"] = annot["time_min_offset"]
                outdf["Class"] = annot["label_class"]
                outdf["Sound type"] = annot["label_subclass"]
                outdf["Software"] = annot["software_name"]
                outdf["Confidence"] = annot["confidence"]
                outdf["Begin Path"] = [
                    os.path.join(x, y) + z
                    for x, y, z in zip(
                        annot["audio_file_dir"],
                        annot["audio_file_name"],
                        annot["audio_file_extension"],
                    )
                ]
                outdf["Begin File"] = [
                    x + y
                    for x, y in zip(
                        annot["audio_file_name"], annot["audio_file_extension"]
                    )
                ]
                outdf = outdf.fillna(0)
                if single_file:
                    outfilename = os.path.join(outdir, outfile)
                else:
                    outfilename = os.path.join(
                        outdir,
                        str(annot["audio_file_name"].iloc[0])
                        + str(annot["audio_file_extension"].iloc[0])
                        + ".chan"
                        + str(annot["audio_channel"].iloc[0])
                        + ".Table.1.selections.txt",
                    )
                outdf.to_csv(
                    outfilename,
                    sep="\t",
                    encoding="utf-8",
                    header=True,
                    columns=cols,
                    index=False,
                )
        else:
            # No annotation => write file with header only
            outfilename = os.path.join(outdir, outfile)
            header = "\t".join(cols)
            f = open(outfilename, "w")
            f.write(header)
            f.close()

    def to_sqlite(self, file):
        """
        Write data to a sqlite database file.

        Write annotations as .sqlite file.

        Parameters
        ----------
        file : str
            Path of the output file (.sqlite) to be written.

        Returns
        -------
        None.

        """
        if file.endswith(".sqlite") is False:
            file = file + ".sqlite"
        self._enforce_dtypes()

        conn = sqlite3.connect(file)
        self.data.to_sql(
            name="detections", con=conn, if_exists="append", index=False
        )
        conn.close()

    def from_sqlite(self, files, table_name="detections", verbose=False):
        """
        Import data from 1 or several sqlite files.

        Load annotation or detection tables from .sqlite files created by the
        method annotation.to_sqlite

        Parameters
        ----------
        files : str, list
            Path of the sqlite file(s) to import. Can be a str if importing a
            single file. Needs to be a list if importing multiple files. If
            'files' is a folder, all files in that folder ending with '.sqlite'
            will be imported.
        table_name : str, optional
            Name of the sql table name containing the annotations. The default
            is 'detections'.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        assert type(files) in (
            str,
            list,
        ), "Input must be of type str (single \
            file or directory) or list (multiple files)"
        files = Annotation._make_list_from_input(
            files, ".sqlite", verbose=verbose
        )
        """Import one or several sqlite files to a Panda datafrane."""
        tmp = []
        for idx, file in enumerate(files):
            conn = sqlite3.connect(file)
            tmp2 = pd.read_sql_query(
                "SELECT * FROM " + table_name,
                conn,
                parse_dates={
                    "entry_date":'ISO8601',
                    "audio_file_start_date":'ISO8601',
                    "time_min_date":'ISO8601',
                    "time_max_date":'ISO8601',
                },
            )
            conn.close()

            tmp.append(tmp2)
        data = pd.concat(tmp, ignore_index=True, sort=False)
        data.reset_index(inplace=True, drop=True)
        self.data = data
        self.check_integrity(verbose=verbose, ignore_frequency_duplicates=True)
        if verbose:
            print(len(self), "annotations imported.")

    def from_pamlab(self, files, verbose=False):
        """
        Import data from 1 or several PAMlab files.

        Load annotation data from .log files generated by the software PAMlab.

        Parameters
        ----------
        files : str, list
            Path of the txt file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple
            files. If 'files' is a folder, all files in that folder ending with
            'annotations.log' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        if type(files) is str:
            if os.path.isdir(files):
                files = ecosound.core.tools.list_files(
                    files,
                    " annotations.log",
                    recursive=False,
                    case_sensitive=True,
                )
                if verbose:
                    print(len(files), "annotation files found.")
        data = Annotation._import_csv_files(files)
        files_timestamp = ecosound.core.tools.filename_to_datetime(
            data["Soundfile"].tolist()
        )
        self.data["audio_file_start_date"] = files_timestamp
        self.data["operator_name"] = data["Operator"]
        self.data["entry_date"] = pd.to_datetime(
            data["Annotation date and time (local)"],
            format="%Y-%m-%d %H:%M:%S.%f",
        )
        self.data["audio_channel"] = data["Channel"]
        self.data["audio_file_name"] = data["Soundfile"].apply(
            lambda x: os.path.splitext(os.path.basename(x))[0]
        )
        self.data["audio_file_dir"] = data["Soundfile"].apply(
            lambda x: os.path.dirname(x)
        )
        self.data["audio_file_extension"] = data["Soundfile"].apply(
            lambda x: os.path.splitext(x)[1]
        )
        self.data["audio_sampling_frequency"] = data["Sampling freq (Hz)"]
        self.data["recorder_type"] = data["Recorder type"]
        self.data["recorder_SN"] = data["Recorder ID"]
        self.data["hydrophone_depth"] = data["Recorder depth"]
        self.data["location_name"] = data["Station"]
        self.data["location_lat"] = data["Latitude (deg)"]
        self.data["location_lon"] = data["Longitude (deg)"]
        self.data["time_min_offset"] = data["Left time (sec)"]
        self.data["time_max_offset"] = data["Right time (sec)"]
        self.data["time_min_date"] = pd.to_datetime(
            self.data["audio_file_start_date"]
            + pd.to_timedelta(self.data["time_min_offset"], unit="s")
        )
        self.data["time_max_date"] = pd.to_datetime(
            self.data["audio_file_start_date"]
            + pd.to_timedelta(self.data["time_max_offset"], unit="s")
        )
        self.data["frequency_min"] = data["Bottom freq (Hz)"]
        self.data["frequency_max"] = data["Top freq (Hz)"]
        self.data["label_class"] = data["Species"]
        self.data["label_subclass"] = data["Call type"]
        self.data["from_detector"] = False
        self.data["software_name"] = "pamlab"
        self.data["uuid"] = self.data.apply(
            lambda _: str(uuid.uuid4()), axis=1
        )
        self.data["duration"] = (
            self.data["time_max_offset"] - self.data["time_min_offset"]
        )
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), "annotations imported.")

    def to_pamlab(
        self, outdir, outfile="PAMlab annotations.log", single_file=False
    ):
        """
        Write data to 1 or several PAMlab files.

        Write annotations as .log files readable by the software PAMlab. Output
        files can be written in a single txt file or in several txt files (one
        per audio recording). In teh latter case, output file names are
        automatically generated based on the audio file's name and the name
        format required by PAMlab.

        Parameters
        ----------
        outdir : str
            Path of the output directory where the Raven files are written.
        outfile : str
            Name of teh output file. Only used is single_file is True. The
            default is 'PAMlab annotations.log'.
        single_file : bool, optional
            If set to True, writes a single output file with all annotations.
            The default is False.

        Returns
        -------
        None.

        """
        if single_file:
            annots = [self.data]
        else:
            annots = [
                pd.DataFrame(y)
                for x, y in self.data.groupby(
                    "audio_file_name", as_index=False
                )
            ]
        for annot in annots:
            annot.reset_index(inplace=True, drop=True)
            cols = [
                "fieldkey:",
                "Soundfile",
                "Channel",
                "Sampling freq (Hz)",
                "Latitude (deg)",
                "Longitude (deg)",
                "Recorder ID",
                "Recorder depth",
                "Start date and time (UTC)",
                "Annotation date and time (local)",
                "Recorder type",
                "Deployment",
                "Station",
                "Operator",
                "Left time (sec)",
                "Right time (sec)",
                "Top freq (Hz)",
                "Bottom freq (Hz)",
                "Species",
                "Call type",
                "rms SPL",
                "SEL",
                "",
                "",
            ]
            outdf = pd.DataFrame(
                {
                    "fieldkey:": 0,
                    "Soundfile": 0,
                    "Channel": 0,
                    "Sampling freq (Hz)": 0,
                    "Latitude (deg)": 0,
                    "Longitude (deg)": 0,
                    "Recorder ID": 0,
                    "Recorder depth": 0,
                    "Start date and time (UTC)": 0,
                    "Annotation date and time (local)": 0,
                    "Recorder type": 0,
                    "Deployment": 0,
                    "Station": 0,
                    "Operator": 0,
                    "Left time (sec)": 0,
                    "Right time (sec)": 0,
                    "Top freq (Hz)": 0,
                    "Bottom freq (Hz)": 0,
                    "Species": "",
                    "Call type": "",
                    "rms SPL": 0,
                    "SEL": 0,
                    "": "",
                    "": "",
                },
                index=list(range(annot.shape[0])),
            )
            outdf["fieldkey:"] = "an:"
            outdf["Species"] = annot["label_class"]
            outdf["Call type"] = annot["label_subclass"]
            outdf["Left time (sec)"] = annot["time_min_offset"]
            outdf["Right time (sec)"] = annot["time_max_offset"]
            outdf["Top freq (Hz)"] = annot["frequency_max"]
            outdf["Bottom freq (Hz)"] = annot["frequency_min"]
            outdf["rms SPL"] = annot["confidence"]
            outdf["Operator"] = annot["operator_name"]
            outdf["Channel"] = annot["audio_channel"]
            outdf["Annotation date and time (local)"] = annot["entry_date"]
            outdf["Sampling freq (Hz)"] = annot["audio_sampling_frequency"]
            outdf["Recorder type"] = annot["recorder_type"]
            outdf["Recorder ID"] = annot["recorder_SN"]
            outdf["Recorder depth"] = annot["hydrophone_depth"]
            outdf["Station"] = annot["location_name"]
            outdf["Latitude (deg)"] = annot["location_lat"]
            outdf["Longitude (deg)"] = annot["location_lon"]
            outdf["Soundfile"] = [
                os.path.join(x, y) + z
                for x, y, z in zip(
                    annot["audio_file_dir"],
                    annot["audio_file_name"],
                    annot["audio_file_extension"],
                )
            ]
            outdf = outdf.fillna(0)
            if single_file:
                outfilename = os.path.join(outdir, outfile)
            else:
                outfilename = os.path.join(
                    outdir,
                    str(annot["audio_file_name"].iloc[0])
                    + str(annot["audio_file_extension"].iloc[0])
                    + " annotations.log",
                )
            outdf.to_csv(
                outfilename,
                sep="\t",
                encoding="utf-8",
                header=True,
                columns=cols,
                index=False,
            )

    def from_parquet(self, file, verbose=False):
        """
        Import data from a Parquet file.

        Load annotations from a .parquet file. This format allows for fast and
        efficient data storage and access.

        Parameters
        ----------
        file : str
            Path of the input parquet file.

        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        self.data = pd.read_parquet(file)
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), "annotations imported.")

    def to_parquet(self, file):
        """
        Write data to a Parquet file.

        Write annotations as .parquet file. This format allows for fast and
        efficient data storage and access.

        Parameters
        ----------
        file : str
            Path of the output directory where the parquet files is written.

        Returns
        -------
        None.

        """
        # make sure the HP SN column are strings
        self.data.hydrophone_SN = self.data.hydrophone_SN.astype(str)
        # save
        self.data.to_parquet(
            file, coerce_timestamps="ms", allow_truncated_timestamps=True
        )

    def from_netcdf(self, files, verbose=False):
        """
        Import data from a netcdf file.

        Load annotations from a .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        files : str
            Path of the nc file to import. Can be a str if importing a single
            file or entire folder. Needs to be a list if importing multiple
            files. If 'files' is a folder, all files in that folder ending with
            '.nc' will be imported.
        verbose : bool, optional
            If set to True, print the summary of the annatation integrity test.
            The default is False.

        Returns
        -------
        None.

        """
        assert type(files) in (
            str,
            list,
        ), "Input must be of type str (single \
            file or directory) or list (multiple files)"
        files = Annotation._make_list_from_input(files, ".nc", verbose=verbose)
        # Import all files to a dataframe
        tmp = []
        for idx, file in enumerate(files):
            dxr = xr.open_dataset(file)
            if dxr.attrs["datatype"] == "Annotation":
                tmp2 = dxr.to_dataframe()
                tmp2.reset_index(inplace=True)
            elif dxr.attrs["datatype"] == "Measurement":
                tmp2 = dxr.to_dataframe()
                tmp2.reset_index(inplace=True)
                tmp2 = tmp2[self.get_fields()]
                warnings.warn(
                    "Importing Measurement data as Annotation >> Not all Measurement data are loaded."
                )
            else:
                raise ValueError(file + "Not an Annotation file.")
            tmp.append(tmp2)

        data = pd.concat(tmp, ignore_index=True, sort=False)
        data.reset_index(inplace=True, drop=True)
        self.data = data
        self.check_integrity(verbose=verbose)
        if verbose:
            print(len(self), "annotations imported.")

    def to_netcdf(self, file):
        """
        Write data to a netcdf file.

        Write annotations as .nc file. This format works well with xarray
        and Dask.

        Parameters
        ----------
        file : str
            Path of the output file (.nc) to be written.

        Returns
        -------
        None.

        """
        if file.endswith(".nc") is False:
            file = file + ".nc"
        self._enforce_dtypes()
        meas = self.data
        meas.set_index("time_min_date", drop=False, inplace=True)
        meas.index.name = "date"
        dxr1 = meas.to_xarray()
        dxr1.attrs["datatype"] = "Annotation"
        dxr1.to_netcdf(file, engine="netcdf4", format="NETCDF4")

    def to_csv(self, file):
        """
        Writes data from the Annoatation object to a csv file.

        Parameters
        ----------
        file : str
            Name and path of the output csv file.

        Returns
        -------
        None.

        """
        self.data.to_csv(file, index=False)

    def insert_values(self, **kwargs):
        """
        Insert constant values for given Annotation fields.

        Fill in entire columns of the annotation dataframe with constant
        values. It is usefull for adding project related informations that may
        not be included in data imported from Raven or PAMlab files (e.g.,
        'location_lat', 'location_lon'). Values can be inserted for several
        annotations fields at a time by setting several keywords. This should
        only be used for filling in static values (i.e., not for variable
        values such as time/frequency boundaries of the annotations). Keywords
        must have the exact same name as the annotation field (see method
        .get_fields). For example: (location_lat=48.6, recorder_type='AMAR')

        Parameters
        ----------
        **kwargs : annotation filed name
            Keyword and value of the annotation field to fill in. Keywords must
            have the exact same name as the annotation field.

        Raises
        ------
        ValueError
            If keyword doesn't match any annotation field name.

        Returns
        -------
        None.

        """
        for key, value in kwargs.items():
            if key in self.data:
                #self.data[key] = value
                self.data.loc[:, key] = value
            else:
                raise ValueError(
                    "The annotation object has no field: " + str(key)
                )

    def insert_metadata(self, deployment_info_file, channel=0):
        """
        Insert metadata infortion to the annotation.

        Uses the Deployment_info_file to fill in the metadata of the annotation
        . The deployment_info_file must be created using the DeploymentInfo
        class from ecosound.core.metadata using DeploymentInfo.write_template.

        Parameters
        ----------
        deployment_info_file : str
            Csv file readable by ecosound.core.meta.DeploymentInfo.read(). It
            contains all the deployment metadata.
        channel : int
            Channel number of the recorder for the metadata to insert.
            Default is 0

        Returns
        -------
        None.

        """
        channel = int(channel)
        dep_info = DeploymentInfo()
        dep_info.read(deployment_info_file)
        self.insert_values(
            UTC_offset=dep_info.data["UTC_offset"].values[channel],
            audio_channel=dep_info.data["audio_channel_number"].values[
                channel
            ],
            audio_sampling_frequency=dep_info.data[
                "sampling_frequency"
            ].values[channel],
            audio_bit_depth=dep_info.data["bit_depth"].values[channel],
            mooring_platform_name=dep_info.data[
                "mooring_platform_name"
            ].values[channel],
            recorder_type=dep_info.data["recorder_type"].values[channel],
            recorder_SN=dep_info.data["recorder_SN"].values[channel],
            hydrophone_model=dep_info.data["hydrophone_model"].values[channel],
            hydrophone_SN=dep_info.data["hydrophone_SN"].values[channel],
            hydrophone_depth=dep_info.data["hydrophone_depth"].values[channel],
            location_name=dep_info.data["location_name"].values[channel],
            location_lat=dep_info.data["location_lat"].values[channel],
            location_lon=dep_info.data["location_lon"].values[channel],
            location_water_depth=dep_info.data["location_water_depth"].values[
                channel
            ],
            deployment_ID=dep_info.data["deployment_ID"].values[channel],
        )

    def insert_metadata_blank(self):
        """
        Insert blank metadata information to the annotation.
        Fill in the blank metadata when the deployment_info_file is not provided

        """
        self.insert_values(audio_channel=0,
                           UTC_offset=0,
                           audio_sampling_frequency=0,
                           audio_bit_depth=int(0),
                           mooring_platform_name='',
                           recorder_type='',
                           recorder_SN='',
                           hydrophone_model='',
                           hydrophone_SN='',
                           hydrophone_depth=0,
                           location_name='',
                           location_lat=0,
                           location_lon=0,
                           location_water_depth=0,
                           deployment_ID='',
                           )

    def filter_overlap_with(
        self,
        annot,
        freq_ovp=True,
        dur_factor_max=None,
        dur_factor_min=None,
        ovlp_ratio_min=None,
        remove_duplicates=False,
        inherit_metadata=False,
        filter_deploymentID=True,
        inplace=False,
    ):
        """
        Filter overalaping annotations.

        Only keep annotations that overlap in time and/or frequency with the
        annotation object "annot".

        Parameters
        ----------
        annot : ecosound.annotation.Annotation object
            Annotation object used to filter the current annotations.
        freq_ovp : bool, optional
            If set to True, filters not only annotations that overlap in time
            but also overlap in frequency. The default is True.
        dur_factor_max : float, optional
            Constraint dictating the maximum duration overlapped
            annotations must not exceed in order to be "kept". Any annotations
            whose duration exceed dur_factor_max*annot.duration are discareded,
            even if they overlap in time/frequency. If set to None, no maximum
            duration constraints are applied. The default is None.
        dur_factor_min : float, optional
            Constraint dictating the minimum duration overlapped
            annotations must exceed in order to be "kept". Any annotations
            whose duration does not exceed dur_factor_min*annot.duration are
            discareded, even if they overlap in time/frequency. If set to None,
            no minimum duration constraints are applied. The default is None.
        ovlp_ratio_min : float, optional
            Constraint dictating the minimum amount (percentage) of overlap in
            time annotations must have in order to be "kept". If set to None,
            no minimum time overlap constraints are applied. The default is
            None.
        remove_duplicates : bool, optional
            If set to True, only selects a single annotation overlaping with
            annotations from the annot object. This is relevant only if several
            annotations overlap with an annotation from the annot object. The
            default is False.
        inherit_metadata : bool, optional
            If set to True, the filtered annotations inherit all the metadata
            information from the matched annotations in the annot object. It
            includes 'label_class', 'label_subclass', 'mooring_platform_name',
            'recorder_type', 'recorder_SN', 'hydrophone_model', 'hydrophone_SN'
            , 'hydrophone_depth', 'location_name', 'location_lat',
            'location_lon', 'location_water_depth', and 'deployment_ID'. The
            default is False.
        filter_deploymentID : bool, optional
            If set to False, doesn't use the deploymentID to match annotations
            together but just the frequency and time offset boundaries of the
            annotations. The default is True.
        inplace : bool, optional
            If set to True, updates the urrent object with the filter results.
            The default is False.

        Returns
        -------
        out_object : ecosound.annotation.Annotation
            Filtered Annotation object.

        """
        stack = []
        det = self.data
        for index, an in annot.data.iterrows():  # for each annotation
            # restrict to the specific deploymnetID of the annotation if file names are not unique
            if filter_deploymentID:
                df = det[det.deployment_ID == an.deployment_ID]
            else:
                df = det
            ## filter detections to same file and deployment ID as the current annotation
            df = df[df.audio_file_name == an.audio_file_name]
            ## check overlap in time first
            if len(df) > 0:
                df = df[
                    (
                        (df.time_min_offset <= an.time_min_offset)
                        & (df.time_max_offset >= an.time_max_offset)
                    )
                    | (  # 1- annot inside detec
                        (df.time_min_offset >= an.time_min_offset)
                        & (df.time_max_offset <= an.time_max_offset)
                    )
                    | (  # 2- detec inside annot
                        (df.time_min_offset < an.time_min_offset)
                        & (df.time_max_offset < an.time_max_offset)
                        & (df.time_max_offset > an.time_min_offset)
                    )
                    | (  # 3- only the end of the detec overlaps with annot
                        (df.time_min_offset > an.time_min_offset)
                        & (df.time_min_offset < an.time_max_offset)
                        & (df.time_max_offset > an.time_max_offset)
                    )  # 4- only the begining of the detec overlaps with annot
                ]
            # then looks at frequency overlap. Can be turned off if freq bounds are not reliable
            if (len(df) > 0) & freq_ovp:
                df = df[
                    (
                        (df.frequency_min <= an.frequency_min)
                        & (df.frequency_max >= an.frequency_max)
                    )
                    | (  # 1- annot inside detec
                        (df.frequency_min >= an.frequency_min)
                        & (df.frequency_max <= an.frequency_max)
                    )
                    | (  # 2- detec inside annot
                        (df.frequency_min < an.frequency_min)
                        & (df.frequency_max < an.frequency_max)
                        & (df.frequency_max > an.frequency_min)
                    )
                    | (  # 3- only the top of the detec overlaps with annot
                        (df.frequency_min > an.frequency_min)
                        & (df.frequency_min < an.frequency_max)
                        & (df.frequency_max > an.frequency_max)
                    )  # 4- only the bottom of the detec overlaps with annot
                ]
            # discard if durations are too different
            if (len(df) > 0) & (dur_factor_max is not None):
                df = df[df.duration < an.duration * dur_factor_max]
            if (len(df) > 0) & (dur_factor_min is not None):
                df = df[df.duration > an.duration * dur_factor_min]

            # discard if they don't overlap enough
            if (len(df) > 0) & (ovlp_ratio_min is not None):
                df_ovlp = (
                    df["time_max_offset"].apply(
                        lambda x: min(x, an.time_max_offset)
                    )
                    - df["time_min_offset"].apply(
                        lambda x: max(x, an.time_min_offset)
                    )
                ) / an.duration
                df = df[df_ovlp >= ovlp_ratio_min]
                df_ovlp = df_ovlp[df_ovlp >= ovlp_ratio_min]

            if (len(df) > 1) & remove_duplicates:
                df_ovlp = (
                              df["time_max_offset"].apply(
                                  lambda x: min(x, an.time_max_offset)
                              )
                              - df["time_min_offset"].apply(
                              lambda x: max(x, an.time_min_offset)
                          )
                          ) / an.duration
                try:
                    df = df.iloc[
                        [df_ovlp.values.argmax()]
                    ]  # pick teh one with max time overlap
                except:
                    print("asas")

            if len(df) > 0:
                if inherit_metadata:
                    df["mooring_platform_name"] = an["mooring_platform_name"]
                    df["recorder_type"] = an["recorder_type"]
                    df["recorder_SN"] = an["recorder_SN"]
                    df["hydrophone_model"] = an["hydrophone_model"]
                    df["hydrophone_SN"] = an["hydrophone_SN"]
                    df["hydrophone_depth"] = an["hydrophone_depth"]
                    df["location_name"] = an["location_name"]
                    df["location_lat"] = an["location_lat"]
                    df["location_lon"] = an["location_lon"]
                    df["location_water_depth"] = an["location_water_depth"]
                    df["deployment_ID"] = an["deployment_ID"]
                    df["label_class"] = an["label_class"]
                    df["label_subclass"] = an["label_subclass"]
                stack.append(df)
        if len(stack) > 0:
            ovlp = pd.concat(stack, ignore_index=True)
        else:
            ovlp = self.data[0:0]
        if inplace:
            self.data = ovlp
            self.check_integrity()
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object.data = ovlp
            out_object.check_integrity()
        return out_object

    def calc_time_aggregate_1D(
        self,
        integration_time="1H",
        resampler="count",
        is_binary=False,
        start_date=None,
        end_date=None,
    ):
        """
        Calculate the 1D time aggregate of annotations.

        Calculate the time aggregate of annotations over the defined
        integration time.

        Parameters
        ----------
        integration_time : str, optional
            Integration time for the aggregate. Uses the pandas offset aliases
            (i.e. '2H'-> 2 hours, '15min'=> 15 minutes, '1D'-> 1 day) see pandas
            documnentation here:https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
            The default is '1H'.
        resampler : str, optional
            Defines method to combine aggregates. Currently only 'count' is
            implemented. The default is 'count'.
        start_date : None, str, optional
            Defines at which date the aggregate should start. If set to None,
            the min. date will be automatically chosen. If str must be in the
            format: yyyy-mm-dd HH:MM:SS. The default is None.
        end_date : None, str, optional
            Defines at which date the aggregate should end. If set to None,
            the max. date will be automatically chosen. If str must be in the
            format: yyyy-mm-dd HH:MM:SS. The default is None.
        is_binary : bool, optional
            If set to True, calculates the aggregates in term on presence (1)
            or absence (0). The default is False.

        Returns
        -------
        data_resamp : Pandas DataFrame
            1D DataFrame with datetime as the index and a 'value' column with the
            result of the aggregates for each time frane.
        """
        # calulate 1D aggreagate
        data = copy.copy(self.data)
        data.set_index("time_min_date", inplace=True)
        data_resamp = Annotation._resample(
            data,
            integration_time=integration_time,
            resampler=resampler,
            start_date=start_date,
            end_date=end_date,
        )
        data_resamp.set_index("datetime", inplace=True)
        if is_binary:
            data_resamp[data_resamp > 0] = 1
        return data_resamp

    def calc_time_aggregate_2D(
        self, integration_time="1H", resampler="count", is_binary=False
    ):
        """
        Calculate the 2D time aggregate of annotations.

        Calculate the time aggregate of annotations for each day and over each
        time of day defined by the integration time (i.e. Time of day vs Date).

        Parameters
        ----------
        integration_time : str, optional
            Integration time for the aggregate. Uses the pandas offset aliases
            (i.e. '2H'-> 2 hours, '15min'=> 15 minutes, '1D'-> 1 day) see pandas
            documnentation here:https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases
            The default is '1H'.
        resampler : str, optional
            Defines method to combine aggregates. Currently only 'count' is
            implemented. The default is 'count'.
        is_binary : bool, optional
            If set to True, calculates the aggregates in term on presence (1)
            or absence (0). The default is False.

        Returns
        -------
        data_resamp : Pandas DataFrame
            2D DataFrame with the time od day as the index and the date as
            columns. The resulting table contains the aggregate of annotations
            for each time of day and date.
        """
        # calulate 1D aggreagate
        data_resamp = self.calc_time_aggregate_1D(
            integration_time=integration_time, is_binary=is_binary
        )
        data_resamp.reset_index(inplace=True)
        data_resamp["date"] = data_resamp["datetime"].dt.date
        data_resamp["time"] = data_resamp["datetime"].dt.time
        # Create 2D matrix
        axis_date = sorted(data_resamp["date"].unique())
        data_grid = pd.pivot_table(
            data_resamp,
            values="value",
            index="time",
            columns="date",
            aggfunc=np.sum,
        )
        data_grid = data_grid.fillna(0)  # replaces NaNs by zeros
        if is_binary:
            data_grid[data_grid > 0] = 1
        return data_grid

    def heatmap(self, **kwargs):
        """
        Display heatmap of annotations (date vs time-of-day).

        Parameters
        ----------
        **kwargs :
            see documentation of ecosound.visualization.annotation_plotter.AnnotHeatmap

        Returns
        -------
        None.

        """
        print(kwargs)
        graph = GrapherFactory("AnnotHeatmap", **kwargs)
        graph.add_data(self)
        graph.show()

    def filter(self, query_str, inplace=False, **kwargs):
        """
        Filter data based on user-defined criteria.

        Uses the pandas dataframe.query method to filter rows of the annotation
        object (in annot.data datafrane). Filtering conditions are defined
        as query string (query_str) indicating the fields and value conditions.
        For example: query_str = 'label_class == "MW" & confidence >= 90'. See
        documentation of pandas.DataFrame.query for more details.

        Parameters
        ----------
        query_str : str
            query string defining annotations fiels and value conditions. For
            exemple: query_str = 'label_class == "MW" & confidence >= 90'.
        inplace : bool, optional
            Whether to modify the DataFrame rather than creating a new one.
            The default is True.

        Returns
        -------
        Annotation object
            Updated with the filtered data.

        """
        # unpack kwargs as separate variables
        for key, val in kwargs.items():
            exec(key + "=val")
        # del key, val
        # filter
        filt = self.data.query(query_str, inplace=inplace)
        # create output obj
        if inplace:
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object.data = filt
            out_object.check_integrity()
        return out_object

    def merge_overlapped(self, time_tolerance_sec=None, inplace=False):

        # add temporary time shift
        if time_tolerance_sec:
            self.data.time_min_offset = self.data.time_min_offset - time_tolerance_sec
            self.data.time_max_offset = self.data.time_max_offset + time_tolerance_sec

        # get index of overlapped annots
        ovlp_idx_list = self._identify_ovlp_annot()

        # merge
        annot2 = Annotation()
        for annot_idx in ovlp_idx_list:
            annot_tmp = Annotation()
            annot_tmp.data = self.data.iloc[[annot_idx[0]]]
            if len(annot_idx)>1:
                t_min = min(self.data.iloc[annot_idx].time_min_offset)
                t_max = max(self.data.iloc[annot_idx].time_max_offset)
                date_min = min(self.data.iloc[annot_idx].time_min_date)
                date_max = max(self.data.iloc[annot_idx].time_max_date)
                f_min = min(self.data.iloc[annot_idx].frequency_min)
                f_max = max(self.data.iloc[annot_idx].frequency_max)
                dur = t_max-t_min-(2*time_tolerance_sec)
                # TO DO
                # apply merge rules for given columns (e.g. SNR, confidence)
                annot_tmp.data.loc[:,'time_min_offset'] = t_min
                annot_tmp.data.loc[:,'time_max_offset'] = t_max
                annot_tmp.data.loc[:,'duration'] = dur
                annot_tmp.data.loc[:,'time_min_date'] = date_min
                annot_tmp.data.loc[:,'time_max_date'] = date_max
                annot_tmp.data.loc[:,'frequency_min'] = f_min
                annot_tmp.data.loc[:,'frequency_max'] = f_max

            # create new dataframe
            annot2 = annot2+annot_tmp

        # remove temporary time shift
        if time_tolerance_sec:
            annot2.data.time_min_offset = annot2.data.time_min_offset + time_tolerance_sec
            annot2.data.time_max_offset = annot2.data.time_max_offset - time_tolerance_sec

        if inplace:
            self.data = annot2.data
            self.check_integrity()
            out_object = None
        else:
            out_object = copy.copy(self)
            out_object.data = annot2.data
            out_object.check_integrity()
        return out_object

    def update_audio_dir(self, new_data_dir, verbose=False):
        """
        Update path of audio files

        Recursively finds the path of the annotations audio files in the folder
        provided in new_data_dir and automatically updates the annotation field
        "audio_file_dir". It is useful when the location of the audio data has
        moved or if using annotations on a different computer.

        Parameters
        ----------
        new_data_dir : str
            Path of the parent directory where the audio files are.
        verbose : bool
            Printprocess logs in command window if set to True. The defaut is
            False.

        Returns
        -------
        None.

        """
        # list name of all audio files in dataset
        dataset_files_list = set(
            self.data["audio_file_dir"]
            + os.path.sep
            + self.data["audio_file_name"]
            + self.data["audio_file_extension"]
        )

        # list extension of all audio files in dataset
        dataset_ext_list = set(
            [os.path.splitext(file)[1] for file in dataset_files_list]
        )
        if verbose:
            print(len(dataset_files_list), " audio files.")

        # list all audio files in new folder (only for the target file extensions)
        new_dir_files_list = []
        for ext in dataset_ext_list:
            new_dir_files_list = (
                new_dir_files_list
                + ecosound.core.tools.list_files(
                    new_data_dir, ext, recursive=True
                )
            )

        # go through each file in dataset and try to find in in new data folder
        missing_files_list = []
        for file in dataset_files_list:
            # if verbose:
            # print(file)
            res = [
                idx
                for idx, new_dir_file in enumerate(new_dir_files_list)
                if re.search(os.path.split(file)[1], new_dir_file)
            ]
            if len(res) == 0:
                missing_files_list.append(file)
            else:
                new_path = os.path.split(new_dir_files_list[res[0]])[0]
                self.data.loc[
                    self.data["audio_file_name"]
                    == os.path.splitext(os.path.split(file)[1])[0],
                    "audio_file_dir",
                ] = new_path

        if len(missing_files_list) > 0:
            warnings.warn(
                str(len(missing_files_list)) + " files could not be found."
            )
            if verbose:
                print("")
                print("List of audio files not found: ")
                print("")
                for ff in missing_files_list:
                    print(ff)
        else:
            if verbose:
                print("Audio paths succesfully updated.")

    def export_spectrograms(
        self,
        out_dir,
        time_buffer_sec=1,
        spectro_unit="samp",
        spetro_nfft=256,
        spetro_frame=256,
        spetro_inc=5,
        freq_min_hz=None,
        freq_max_hz=None,
        sanpling_rate_hz=None,
        filter_order=8,
        filter_type="iir",
        fig_size=(15, 10),
        deployment_subfolders=False,
        date_subfolders=False,
        date_subfolders_format="%Y-%m-%d",
        file_subfolder=False,
        file_name_field="uuid",
        file_prefix_field=None,
        channel=None,
        colormap="viridis",
        save_wav=False,
    ):

        # define the different class names and create separate folders
        if os.path.isdir(out_dir) == False:
            os.mkdir(out_dir)
        labels = list(set(self.data["label_class"]))
        # labels.reverse()

        # initialize spectrogram
        Spectro = Spectrogram(
            spetro_frame,
            "hann",
            spetro_nfft,
            spetro_inc,
            sanpling_rate_hz,
            unit=spectro_unit,
        )

        # loop through each class_labels
        for label in labels:
            # print(label)
            current_dir = os.path.join(out_dir, label)
            if os.path.isdir(current_dir) == False:
                os.mkdir(current_dir)
            annot_sp = self.data[self.data["label_class"] == label]

            # loop through is annot for that class label
            for idx, annot in tqdm(
                annot_sp.iterrows(),
                desc=label,
                leave=True,
                position=0,
                miniters=1,
                total=len(annot_sp),
                colour="green",
            ):

                try:
                    # annot = annot_sp.iloc[1969]
                    # idx = 1969
                    # output file name
                    F = self._convert_to_str(annot[file_name_field])

                    # create subfolder for each deployment and each day if option selected
                    if deployment_subfolders:
                        current_dir2 = os.path.join(
                            current_dir, str(annot.deployment_ID)
                        )
                    else:
                        current_dir2 = current_dir
                    if date_subfolders:
                        current_date = annot.time_min_date.strftime(date_subfolders_format)
                        current_dir2 = os.path.join(current_dir2, current_date)

                    if file_subfolder:
                        current_file = annot.audio_file_name
                        current_dir2 = os.path.join(current_dir2, current_file)

                    if os.path.isdir(current_dir2) == False:
                        os.mkdir(current_dir2)

                    # only if file doesn't exist already
                    if os.path.isfile(os.path.join(current_dir2, F)) == False:
                        # print("Processing file", F)

                        # Load info from audio file
                        audio_data = Sound(
                            os.path.join(
                                annot["audio_file_dir"], annot["audio_file_name"]
                            )
                            + annot["audio_file_extension"]
                        )

                        # define start/stop times +/- buffer
                        t1 = annot.time_min_offset - time_buffer_sec
                        if t1 <= 0:
                            t1 = 0
                        t2 = annot.time_max_offset + time_buffer_sec
                        if t2 > audio_data.file_duration_sec:
                            t2 = audio_data.file_duration_sec
                        duration = t2 - t1

                        # load audio data
                        if channel != None:
                            chan = int(channel)
                        else:
                            chan = annot["audio_channel"] - 1
                        audio_data.read(
                            channel=chan,
                            chunk=[t1, t2],
                            unit="sec",
                            detrend=True,
                        )

                        # decimate
                        audio_data.decimate(sanpling_rate_hz)

                        # normalize
                        audio_data.normalize()

                        # compute spectrogram
                        _ = Spectro.compute(audio_data, dB=True, use_dask=False)

                        # crop if needed
                        if freq_min_hz != None or freq_max_hz != None:
                            Spectro.crop(
                                frequency_min=freq_min_hz,
                                frequency_max=freq_max_hz,
                                inplace=True,
                            )

                        # display/save spectrogram as image file
                        graph = GrapherFactory(
                            "SoundPlotter",
                            title=annot["audio_file_name"]
                            + " - "
                            + str(abs(round(annot["time_min_offset"], 2))),
                            fig_size=fig_size,
                            colormap=colormap,
                        )
                        # crop plot if needed
                        if freq_min_hz != None:
                            graph.frequency_min = freq_min_hz
                        if freq_max_hz != None:
                            graph.frequency_max = freq_max_hz

                        # output file name
                        graph.add_data(Spectro)
                        if file_prefix_field:
                            prefix = self._convert_to_str(annot[file_prefix_field])
                            full_out_file = os.path.join(
                                current_dir2, prefix + "_" + F
                            )
                        else:
                            full_out_file = os.path.join(current_dir2, F)

                        graph.to_file(full_out_file + ".png")

                        if save_wav:
                            audio_data.write(full_out_file + ".wav")

                        # graph.show()

                        # if params["spetro_on_npy"]:
                        #    np.save(os.path.splitext(outfilename)[0] + ".npy", S)
                        # annot_unique_id += 1
                    # else:
                    # print("file ", F, " already processed.")
                except Exception as error:
                    print("An error occurred:", type(error).__name__, "–", error)

    def get_labels_class(self):
        """
        Get all the unique class labels of the annotations.

        Returns
        -------
        classes : list
            List of unique class labels.

        """
        if len(self.data) > 0:
            classes = list(self.data["label_class"].unique())
        else:
            classes = []
        return classes

    def get_labels_subclass(self):
        """
        Get all the unique subclass labels of the annotations.

        Returns
        -------
        classes : list
            List of unique subclass labels.

        """
        if len(self.data) > 0:
            subclasses = list(self.data["label_subclass"].unique())
        else:
            subclasses = []
        return subclasses

    def get_fields(self):
        """
        Get all the annotations fields.

        Returns
        -------
        classes : list
            List of annotation fields.

        """
        return list(self.data.columns)

    def summary(self, rows="deployment_ID", columns="label_class"):
        """
        Produce a summary table of the number of annotations.

        Create a pivot table summarizing the number of annotations for each
        deployment and each label class. The optional arguments 'rows' and
        'columns' can be used to change the fields of the annotations to be
        displayed in the table.

        Parameters
        ----------
        rows : 'str', optional
            Name of the annotation field for the rows of the table. The default
            is 'deployment_ID'.
        columns : 'str', optional
            Name of the annotation field for the columns of the table. The
            default default is 'label_class'.

        Returns
        -------
        summary : pandas DataFrame
            Pivot table with the number of annotations in each category.

        """
        summary = self.data.pivot_table(
            index=rows, columns=columns, aggfunc="size", fill_value=0
        )
        # Add a "Total" row and column
        summary.loc["Total"] = summary.sum()
        summary["Total"] = summary.sum(axis=1)
        return summary

    def _identify_ovlp_annot(self):
        stack = []
        data = self.data
        files_list = list(set(data.audio_file_name))
        for file in files_list:  # for each audio file
            file_data = data.query('audio_file_name==@file')  # data for a single file
            while len(file_data) > 0:
                t1 = file_data.iloc[0].time_min_offset
                t2 = file_data.iloc[0].time_max_offset
                # stack index + update dataframe
                tmp = []
                index_id = file_data.iloc[0].name
                tmp.append(index_id)
                file_data = file_data.drop([index_id], axis=0)  # delete annot alreday stacked in tmp
                while True:
                    # find other annot overlaping with curent annot
                    ovlp = file_data[
                        (
                            (file_data.time_min_offset <= t1)
                            & (file_data.time_max_offset >= t2)
                        )
                        | (  # 1- annot inside detec
                            (file_data.time_min_offset >= t1)
                            & (file_data.time_max_offset <= t2)
                        )
                        | (  # 2- detec inside annot
                            (file_data.time_min_offset < t1)
                            & (file_data.time_max_offset < t2)
                            & (file_data.time_max_offset > t1)
                        )
                        | (  # 3- only the end of the detec overlaps with annot
                            (file_data.time_min_offset > t1)
                            & (file_data.time_min_offset < t2)
                            & (file_data.time_max_offset > t2)
                        )  # 4- only the begining of the detec overlaps with annot
                        ]
                    if len(ovlp) == 0:  # no overlap
                        stack.append(tmp)
                        break
                    elif len(ovlp) > 0:  # 1 of more overlaps
                        index_ids = list(ovlp.index.values)
                        for index_id in index_ids:
                            tmp.append(index_id)
                            t1 = min([t1, ovlp.time_min_offset.values[0]])
                            t2 = max([t2, ovlp.time_max_offset.values[0]])
                            file_data = file_data.drop([index_id], axis=0)  # delete annot alreday stacked in tmp
        #print('done')
        # Sanity check
        annot_count = sum([len(ovlp) for ovlp in stack])
        if annot_count != len(data):
            print('overlapped annotation identified do not add up to the total number of annotations')
        return stack

    @staticmethod
    def _resample(
        data,
        integration_time="1H",
        resampler="count",
        start_date=None,
        end_date=None,
    ):

        # define start time of the aggregate
        if start_date is None:
            start_date = min(data.index)
        elif type(start_date) is str:
            start_date = pd.Timestamp(start_date)

        # define end time of the aggregate
        if end_date is None:
            end_date = max(data.index)
        elif type(end_date) is str:
            end_date = pd.Timestamp(end_date)

        start_date = start_date.to_period(integration_time).to_timestamp()
        end_date = end_date.to_period(integration_time).to_timestamp()

        t_index = pd.DatetimeIndex(
            pd.date_range(
                start=start_date,
                end=end_date,
                freq=integration_time,
            )
        )
        if resampler == "count":
            # Lmean = data.resample(integration_time, loffset=None, label='left').apply(count)
            data_new = data.resample(
                integration_time,
                #loffset=None,
                origin="start_day",
                label="left",
            ).count()

        data_out = pd.DataFrame(
            {"datetime": data_new.index, "value": data_new["uuid"]}
        )
        data_out = data_out.reindex(t_index).fillna(0)
        data_out["datetime"] = data_out.index
        data_out.reset_index(drop=True, inplace=True)
        return data_out

    def _enforce_dtypes(self):
        self.data = self.data.astype(
            {
                "uuid": "str",
                "from_detector": "bool",  # True, False
                "software_name": "str",
                "software_version": "str",
                "operator_name": "str",
                "UTC_offset": "float",
                "entry_date": "datetime64[ns]",
                "audio_channel": "int",
                "audio_file_name": "str",
                "audio_file_dir": "str",
                "audio_file_extension": "str",
                "audio_file_start_date": "datetime64[ns]",
                "audio_sampling_frequency": "int",
                "audio_bit_depth": "int",
                "mooring_platform_name": "str",
                "recorder_type": "str",
                "recorder_SN": "str",
                "hydrophone_model": "str",
                "hydrophone_SN": "str",
                "hydrophone_depth": "float",
                "location_name": "str",
                "location_lat": "float",
                "location_lon": "float",
                "location_water_depth": "float",
                "deployment_ID": "str",
                "frequency_min": "float",
                "frequency_max": "float",
                "time_min_offset": "float",
                "time_max_offset": "float",
                "time_min_date": "datetime64[ns]",
                "time_max_date": "datetime64[ns]",
                "duration": "float",
                "label_class": "str",
                "label_subclass": "str",
                "confidence": "float",
            }
        )

    @staticmethod
    @ecosound.core.decorators.listinput
    def _import_csv_files(files):
        """Import one or several text files with header to a Panda datafrane."""
        assert type(files) in (
            str,
            list,
        ), "Input must be of type str (single \
            file) or list (multiple files)"
        # Import all files to a dataframe
        for idx, file in enumerate(files):
            # Extract header first due to formating issues in PAMlab files
            header = pd.read_csv(file, delimiter="\t", header=None, nrows=1)
            headerLength = header.shape[1]
            # Get all data and only keep values correpsonding to header labels
            # tmp = pd.read_csv(file,
            #                   delimiter='\t',
            #                   header=None,
            #                   #header=True,
            #                   #skiprows=1,
            #                   na_values=None,
            #                   )
            # tmp = tmp.iloc[:, 0:headerLength]
            # Put header back
            # tmp = tmp.set_axis(list(header.values[0]), axis=1, inplace=False)
            tmp = pd.read_csv(file, delimiter="\t", na_values=None)
            if idx == 0:
                data = tmp
            else:
                data = pd.concat([data, tmp], ignore_index=True, sort=False)
        return data

    @staticmethod
    def _make_list_from_input(files, file_ext, verbose=True):
        if type(files) is str:
            if os.path.isdir(files):
                files = ecosound.core.tools.list_files(
                    files,
                    file_ext,
                    recursive=False,
                    case_sensitive=True,
                )
                if verbose:
                    print(len(files), "files found.")
            else:
                files = [files]
        return files

    def _convert_to_str(self, value):
        if type(value) is str:
            F = value
        elif type(value) is float:
            if value < 0:
                F = "minus-" + str(abs(round(value, 2)))
            else:
                F = str(round(value, 2))
        elif type(value) is pd.Timestamp:
            if self.data.UTC_offset.iloc[0] >= 0:
                sign_str = "+"
            else:
                sign_str = "-"
            tz_str = datetime.datetime(
                year=1,
                month=1,
                day=1,
                hour=int(abs(self.data.UTC_offset.iloc[0])),
            ).strftime("%H%M")

            F = value.strftime("%Y%m%dT%H%M%S.%f") + sign_str + tz_str
        elif np.isnan(value):
            F = "nan"
        return F

    def __add__(self, other):
        """Concatenate data from several annotation objects."""
        assert (
            type(other) is ecosound.core.annotation.Annotation
        ), "Object type not \
            supported. Can only concatenate Annotation objects together."
        self._enforce_dtypes()
        other._enforce_dtypes()
        self.data = pd.concat(
            [self.data, other.data], ignore_index=True, sort=False
        )
        return self

    def __repr__(self):
        """Return the type of object."""
        return f"{self.__class__.__name__} object (" f"{len(self.data)})"

    def __str__(self):
        """Return string when used with print of str."""
        return f"{len(self.data)} annotation(s)"

    def __len__(self):
        """Return number of annotations."""
        return len(self.data)



# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 21:21:40 2020

@author: xavier.mouy
"""
import json
import re
from datetime import datetime
import ecosound.core.decorators
import numpy as np
from scipy import interpolate
from scipy.signal import argrelmax
import os, sys
from numba import njit
import pkg_resources
import yaml


def read_json(file):
    """Load JSON file as dict."""
    with open(file, "r") as read_file:
        data = json.load(read_file)
    return data


def read_yaml(file):
    """
    Load config file.

    Parameters
    ----------
    file : str
        Path of the yaml file with all the parameters.

    Returns
    -------
    config : dict
        Parsed parameters.

    """
    # Loads config  files
    yaml_file = open(file)
    config = yaml.load(yaml_file, Loader=yaml.FullLoader)
    return config


@ecosound.core.decorators.listinput
def filename_to_datetime(files):
    """Extract date from a list of str of filenames."""
    current_dir = os.path.dirname(os.path.realpath(__file__))
    patterns = read_json(os.path.join(current_dir, r"timestamp_formats.json"))

    # stream = pkg_resources.resource_stream(__name__, 'core/timestamp_formats.json')
    # patterns = read_json(os.path.join(stream)

    regex_string = "|".join(
        [pattern["string_pattern"] for pattern in patterns]
    )
    time_formats = [pattern["time_format"] for pattern in patterns]
    timestamps = [None] * len(files)
    p = re.compile(regex_string)
    for idx, file in enumerate(files):
        datestr = p.search(file)
        for time_format in time_formats:
            ok_flag = False
            try:
                timestamps[idx] = datetime.strptime(datestr[0], time_format)
                ok_flag = True
            except:
                ok_flag = False
            if ok_flag is True:
                break
        if ok_flag is False:
            raise ValueError("Time format not recognized:" + file)
    return timestamps


# @njit
def normalize_vector(vec):
    """
    Normalize amplitude of vector.
    """
    # vec = vec+abs(min(vec))
    # normVec = vec/max(vec)
    # normVec = (normVec - 0.5)*2
    vec = vec - np.mean(vec)
    normVec = vec / max(vec)
    return normVec


# @njit
def tighten_signal_limits(signal, energy_percentage):
    """
    Tighten signal limits

    Redefine start and stop samples to have "energy_percentage" of the original
    signal. Returns a list with the new start and stop sample indices.

    """
    cumul_energy = np.cumsum(np.square(signal))
    cumul_energy = cumul_energy / max(cumul_energy)
    percentage_begining = (1 - (energy_percentage / 100)) / 2
    percentage_end = 1 - percentage_begining
    chunk = [
        np.nonzero(cumul_energy > percentage_begining)[0][0],
        np.nonzero(cumul_energy > percentage_end)[0][0],
    ]
    return chunk


def tighten_signal_limits_peak(signal, percentage_max_energy):
    """
    Tighten signal limits

    Redefine start and stop samples to have "energy_percentage" of the original
    signal. Returns a list with the new start and stop sample indices.

    small values of percentage_max_energy -> tighter signal

    """
    squared_signal = np.square(signal)
    norm_factor = sum(squared_signal)
    squared_signal_normalized = squared_signal / norm_factor
    sort_idx = np.argsort(-squared_signal_normalized)
    sort_val = squared_signal_normalized[sort_idx]
    sort_val_cum = np.cumsum(sort_val)
    id_limit = np.where(sort_val_cum > (percentage_max_energy / 100))
    id_limit = id_limit[0][0]
    min_idx_limit = np.min(sort_idx[0:id_limit])
    max_idx_limit = np.max(sort_idx[0:id_limit])
    chunk = [min_idx_limit, max_idx_limit]

    return chunk


def resample_1D_array(x, y, resolution, kind="linear"):
    """
    Interpolate values of coordinates x and y with a given resolution.
    Default uisn linear interpolation.
    """
    f = interpolate.interp1d(x, y, kind=kind, fill_value="extrapolate")
    xnew = np.arange(x[0], x[-1] + resolution, resolution)
    ynew = f(xnew)
    return xnew, ynew


@njit
def entropy(array_1d, apply_square=False):
    """
    Aggregate (SHannon's) entropy as defined in the Raven manual
    apply_square = True, suqares the array value before calculation.
    """
    if apply_square:
        array_1d = np.square(array_1d)
    values_sum = np.sum(array_1d)
    H = 0
    for value in array_1d:
        ratio = value / values_sum
        if ratio <= 0:
            H += 0
        else:
            H += ratio * np.log2(ratio)
    return H


@njit
def derivative_1d(array, order=1):
    """
    Derivative of order "order" of a 1D array. Subtract the element i+1 to i.
    If order > 1, the process is repeated iteratively "order" times. Note that
    the resulting array is shorter than the original array by "order" elements.

    """
    for n in range(0, order, 1):
        array = np.subtract(array[1:], array[0:-1])
    return array


def list_files(indir, suffix, case_sensitive=True, recursive=False):
    """
    List files in folder whose name ends with a given suffix/extension.

    Parameters
    ----------
    indir : str
        Path of the folder to search.
    suffix : str
        Suffix of the filename.
    case_sensitive : bool, optional
        If set to True, search using case sensitive filenames. The default is
        True.
    recursive : bool, optional
        If set to True, search in parent folder but also in all its subfolders.
        The default is False.

    Returns
    -------
    files_list : list
        List of strings with full path of files found.

    """
    if os.path.isdir(indir):
        files_list = []
        if case_sensitive is False:
            suffix = suffix.lower()
        if recursive:  # scans subfolders recursively
            for root, dirs, files in os.walk(indir):
                for file in files:
                    if case_sensitive is False:
                        file = file.lower()
                    if file.endswith(suffix):
                        files_list.append(os.path.join(root, file))
                        # print(os.path.join(root, file))
        else:  # only scans parent folder
            for file in os.listdir(indir):
                if case_sensitive is False:
                    file = file.lower()
                if file.endswith(suffix):
                    files_list.append(os.path.join(indir, file))
                    # print(os.path.join(indir, file))
    else:
        raise Exception("The indir folder given does not exist.")
    return files_list


@njit
def find_peaks(array, troughs=False):
    """
    Find peaks or troughs in an 1-D array.

    Parameters
    ----------
    array : numpy array or list
        1-dimensional array.
    troughs : bool, optional
        If set to True, finds troughs instead of peaks in the input array.
        The default is False.

    Returns
    -------
    x : list
        Indices of peaks or troughs
    y : list
        Values of peaks or troughs

    """

    x = [
        0,
    ]
    y = [
        array[0],
    ]
    for k in range(1, len(array) - 1):
        if troughs:
            if (np.sign(array[k] - array[k - 1]) == -1) and (
                (np.sign(array[k] - array[k + 1])) == -1
            ):
                x.append(k)
                y.append(array[k])
        else:
            if (np.sign(array[k] - array[k - 1]) == 1) and (
                np.sign(array[k] - array[k + 1]) == 1
            ):
                x.append(k)
                y.append(array[k])
    return x, y


def envelope(array, interp="cubic"):
    # initialize output arrays
    env_high = np.zeros(array.shape)
    env_low = np.zeros(array.shape)
    # Prepend the first value of (s) to the interpolating values. This forces
    # the model to use the same starting point for both the upper and lower
    # envelope models.
    u_x = [
        0,
    ]
    u_y = [
        array[0],
    ]
    l_x = [
        0,
    ]
    l_y = [
        array[0],
    ]
    # Detect peaks and troughs
    l_x, l_y = find_peaks(array, troughs=True)
    u_x, u_y = find_peaks(array, troughs=False)
    # Append the last value of (s) to the interpolating values. This forces the
    # model to use the same ending point for both the upper and lower envelope
    # models.
    u_x.append(len(array) - 1)
    u_y.append(array[-1])
    l_x.append(len(array) - 1)
    l_y.append(array[-1])

    # Interpolate between peaks/troughs
    u_p = interpolate.interp1d(
        u_x, u_y, kind=interp, bounds_error=False, fill_value=0.0
    )
    l_p = interpolate.interp1d(
        l_x, l_y, kind=interp, bounds_error=False, fill_value=0.0
    )
    for k in range(0, len(array)):
        env_high[k] = u_p(k)
        env_low[k] = l_p(k)
    return env_high, env_low



# -*- coding: utf-8 -*-
"""
Created on Wed Oct 19 09:48:18 2022

@author: xavier.mouy
"""

from ecosound.core.annotation import Annotation
from ecosound.core.measurement import Measurement
from matplotlib.patches import Rectangle
from matplotlib.collections import PatchCollection
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import os
from tqdm import tqdm


class PRF:
    def __init__(self, mode):
        pass

    @staticmethod
    def count(
        annot=None,
        detec=None,
        out_dir=None,
        target_class=None,
        files_to_use="both",  # 'detec', 'annot', 'both', list
        thresholds=np.arange(0, 1.05, 0.05),
        date_min=None,
        date_max=None,
        freq_ovp=True,
        dur_factor_max=None,
        dur_factor_min=None,
        ovlp_ratio_min=None,
        remove_duplicates=False,
        inherit_metadata=False,
        filter_deploymentID=False,
        do_plot=False,
        F_beta=1,
    ):

        # create subfolder for figures is plot option is ON
        if do_plot:
            out_dir_figs = os.path.join(out_dir, 'TP-FN-FP-plots')
            if os.path.isdir(out_dir_figs) == False:
                os.mkdir(out_dir_figs)

        # filter dates (if dataset partially annotated)
        if date_min:
            annot.filter("time_min_date >= '" + date_min + "'", inplace=True)
            detec.filter("time_min_date >= '" + date_min + "'", inplace=True)
        if date_max:
            annot.filter("time_max_date <= '" + date_max + "'", inplace=True)
            detec.filter("time_max_date <= '" + date_max + "'", inplace=True)

        # filter to the class of interest
        annot.filter('label_class == "' + target_class + '"', inplace=True)
        detec.filter('label_class == "' + target_class + '"', inplace=True)

        # Define list of files to use for the performance evaluation.
        if (
            files_to_use == "detec"
        ):  # case 1: all files with detections are used
            files_list = list(set(detec.data.audio_file_name))
        elif (
            files_to_use == "annot"
        ):  # case 2: all files with annotations are used
            files_list = list(set(annot.data.audio_file_name))
        elif (
            files_to_use == "both"
        ):  # case 3: all files with annotations or detections are used
            files_list1 = list(set(annot.data.audio_file_name))
            files_list2 = list(set(detec.data.audio_file_name))
            files_list = list(set(files_list1 + files_list2))

        elif (
            type(files_to_use) is list
        ):  # case 4: only files provided by the user are used
            files_list = files_to_use
        files_list.sort()

        # filter annotations with selected files to use
        annot.filter(
            "audio_file_name in @files_list",
            files_list=files_list,
            inplace=True,
        )

        # filter detections with selected files to use
        detec.filter(
            "audio_file_name in @files_list",
            files_list=files_list,
            inplace=True,
        )

        # loop through thresholds
        for th_idx, threshold in enumerate(thresholds):
            print("Threshold value: ", threshold)
            # filter detections for that threshold value
            detec_conf = detec.filter(
                "confidence >= " + str(threshold), inplace=False
            )
            # init
            FP = np.zeros(len(files_list))
            TP = np.zeros(len(files_list))
            FN = np.zeros(len(files_list))
            # loop through each file
            # for idx, file in enumerate(
            #     tqdm(files_list, desc="File", leave=True, miniters=1, colour="red")
            # ):
            # for idx, file in enumerate(files_list):
            for idx, file in enumerate(
                tqdm(
                    files_list,
                    desc="Progress",
                    leave=True,
                    miniters=1,
                    colour="green",
                )
            ):
                # print(idx)
                # filter to only keep data from this file
                annot_tmp = annot.filter("audio_file_name == '" + file + "'")
                detec_tmp = detec_conf.filter(
                    "audio_file_name=='" + file + "'", inplace=False
                )

                # count FP, TP, FN:
                if (
                    len(annot_tmp) == 0
                ):  # if no annotations -> FP = nb of detections
                    FP[idx] = len(detec_tmp)
                elif (
                    len(detec_tmp) == 0
                ):  # if no detections -> FN = nb of annotations
                    FN[idx] = len(annot_tmp)
                else:
                    ovlp = annot_tmp.filter_overlap_with(
                        detec_tmp,
                        freq_ovp=freq_ovp,
                        dur_factor_max=dur_factor_max,
                        dur_factor_min=dur_factor_min,
                        ovlp_ratio_min=ovlp_ratio_min,
                        remove_duplicates=remove_duplicates,
                        inherit_metadata=inherit_metadata,
                        filter_deploymentID=filter_deploymentID,
                        inplace=False,
                    )
                    FN[idx] = len(annot_tmp) - len(ovlp)
                    TP[idx] = len(ovlp)
                    FP[idx] = len(detec_tmp) - len(ovlp)

                # Sanity check
                if FP[idx] + TP[idx] != len(detec_tmp):
                    raise Exception(
                        "FP and TP don't add up to the total number of detections"
                    )
                elif TP[idx] + FN[idx] != len(annot_tmp):
                    raise Exception(
                        "FP and FN don't add up to the total number of annotations"
                    )

                # plot annot and detec boxes
                if do_plot:
                    PRF._plot_annot_boxes(
                        [annot_tmp, detec_tmp],
                        out_dir_figs,
                        threshold,
                        file,
                        line_width=2,
                        colors=["blue", "red"],
                        labels=["Annotations", "Detections"],
                        title=file
                        + "\n TP:"
                        + str(TP[idx])
                        + " - FP: "
                        + str(FP[idx])
                        + " - FN: "
                        + str(FN[idx]),
                    )
            # create df for each file with threshold, TP, FP, FN
            tmp = pd.DataFrame(
                {
                    "file": files_list,
                    "threshold": threshold,
                    "TP": TP,
                    "FP": FP,
                    "FN": FN,
                }
            )
            if th_idx == 0:
                performance_per_file_count = tmp
            else:
                performance_per_file_count = pd.concat(
                    [performance_per_file_count, tmp], ignore_index=True
                )

            # Calculate P, R, and F for that confidence threshold
            perf_tmp = performance_per_file_count.query(
                "threshold ==" + str(threshold)
            )
            TP_th = perf_tmp["TP"].sum()
            FP_th = perf_tmp["FP"].sum()
            FN_th = perf_tmp["FN"].sum()
            R_th = TP_th / (TP_th + FN_th)
            if (TP_th + FP_th) == 0:
                P_th = 1
            else:
                P_th = TP_th / (TP_th + FP_th)
            F_th = (1 + F_beta**2) * ((P_th * R_th) / ((F_beta**2 * P_th) + R_th))
            tmp_PRF = pd.DataFrame(
                {
                    "threshold": [threshold],
                    "TP": [TP_th],
                    "FP": [FP_th],
                    "FN": [FN_th],
                    "R": [R_th],
                    "P": [P_th],
                    "F": [F_th],
                }
            )
            # Sanity check
            if FP_th + TP_th != len(detec_conf):
                raise Exception(
                    "FP and TP don't add up to the total number of detections"
                )
            elif TP_th + FN_th != len(annot):
                raise Exception(
                    "FP and FN don't add up to the total number of annotations"
                )

            if th_idx == 0:
                performance_PRF = tmp_PRF
            else:
                performance_PRF = pd.concat(
                    [performance_PRF, tmp_PRF], ignore_index=True
                )
        # plot PRF curves
        fig, ax = plt.subplots(1, 2)
        # ax[0].axis('equal')
        ax[0].plot(performance_PRF["P"], performance_PRF["R"], "k")
        ax[0].set_xlabel("Precision")
        ax[0].set_ylabel("Recall")
        ax[0].set(xlim=(0, 1), ylim=(0, 1))
        ax[0].grid()
        ax[0].set_aspect("equal", "box")
        fig.tight_layout()

        # ax[1].axis('equal')
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["P"],
            ":k",
            label="Precision",
        )
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["R"],
            "--k",
            label="Recall",
        )
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["F"],
            "k",
            label="$F_" + str(F_beta) + "$-score",
        )
        ax[1].set_xlabel("Threshold")
        ax[1].set_ylabel("Score")
        ax[1].set(xlim=[0, 1], ylim=[0, 1])
        ax[1].grid()
        ax[1].legend()
        ax[1].set_aspect("equal", "box")
        fig.set_size_inches(8, 4)
        fig.tight_layout()
        fig.savefig(os.path.join(out_dir, "Performance_graph.png"))

        # save CSV files
        performance_PRF.to_csv(
            os.path.join(out_dir, "Performance_full_dataset.csv"), index=False
        )
        performance_per_file_count.to_csv(
            os.path.join(out_dir, "Performance_per_file.csv"), index=False
        )

    @staticmethod
    def presence(
        annot=None,
        detec=None,
        out_dir=None,
        target_class=None,
        files_to_use="both",  # 'detec', 'annot', 'both', list, None
        thresholds=np.arange(0, 1.05, 0.05),
        date_min=None,
        date_max=None,
        F_beta=1,
        integration_time="1D",
        min_detec_nb=1,
    ):

        # # load ground truth data
        # annot = Annotation()
        # annot.from_netcdf(annot_file)

        # # annot = Annotation()
        # # annot.from_raven(
        # #     annot_file,
        # #     class_header="Call Type",
        # #     subclass_header="Overlap",
        # #     verbose=True,
        # # )

        # # load destections
        # detec = Annotation()
        # detec.from_sqlite(detec_file)

        # filter dates (if dataset partially annotated)
        if date_min:
            annot.filter("time_min_date >= '" + date_min + "'", inplace=True)
            detec.filter("time_min_date >= '" + date_min + "'", inplace=True)
        if date_max:
            annot.filter("time_max_date <= '" + date_max + "'", inplace=True)
            detec.filter("time_max_date <= '" + date_max + "'", inplace=True)

        # filter to the class of interest
        annot.filter('label_class == "' + target_class + '"', inplace=True)
        detec.filter('label_class == "' + target_class + '"', inplace=True)

        # Define list of files to use for the performance evaluation.
        if files_to_use:
            if (
                files_to_use == "detec"
            ):  # case 1: all files with detections are used
                files_list = list(set(detec.data.audio_file_name))
            elif (
                files_to_use == "annot"
            ):  # case 2: all files with annotations are used
                files_list = list(set(annot.data.audio_file_name))
            elif (
                files_to_use == "both"
            ):  # case 3: all files with annotations or detections are used
                files_list1 = list(set(annot.data.audio_file_name))
                files_list2 = list(set(detec.data.audio_file_name))
                files_list = list(set(files_list1 + files_list2))

            elif (
                type(files_to_use) is list
            ):  # case 4: only files provided by the user are used
                files_list = files_to_use
            files_list.sort()

            # filter annotations with selected files to use
            annot.filter(
                "audio_file_name in @files_list",
                files_list=files_list,
                inplace=True,
            )

            # filter detections with selected files to use
            detec.filter(
                "audio_file_name in @files_list",
                files_list=files_list,
                inplace=True,
            )

        # Create annootation aggregate
        aggr_min_date = min(
            min(annot.data["time_max_date"]), min(detec.data["time_max_date"])
        )
        aggr_max_date = max(
            max(annot.data["time_max_date"]), max(detec.data["time_max_date"])
        )
        annot_aggr = annot.calc_time_aggregate_1D(
            integration_time=integration_time,
            is_binary=False,
            start_date=aggr_min_date,
            end_date=aggr_max_date,
        )

        dates_list = annot_aggr.index

        # loop through thresholds
        # for th_idx, threshold in enumerate(
        #     tqdm(thresholds, desc="Progress", leave=True, miniters=1, colour="green")
        # ):
        for th_idx, threshold in enumerate(thresholds):
            print("Threshold value: ", threshold)
            # filter detections for that threshold value
            detec_conf = detec.filter(
                "confidence >= " + str(threshold), inplace=False
            )

            # Create detection aggregate
            detec_aggr = detec_conf.calc_time_aggregate_1D(
                integration_time=integration_time,
                is_binary=False,
                start_date=aggr_min_date,
                end_date=aggr_max_date,
            )

            # init
            FP = np.zeros(len(dates_list))
            TP = np.zeros(len(dates_list))
            FN = np.zeros(len(dates_list))
            # loop through each file
            # for idx, file in enumerate(
            #     tqdm(files_list, desc="File", leave=True, miniters=1, colour="red")
            # ):
            # for idx, file in enumerate(files_list):
            for idx, date in enumerate(
                tqdm(
                    dates_list,
                    desc="Progress",
                    leave=True,
                    miniters=1,
                    colour="green",
                )
            ):
                # filter to only keep data from this time frame
                try:
                    annot_tmp = annot_aggr.loc[date].value
                    detec_tmp = detec_aggr.loc[date].value
                except:
                    print("stop")

                # apply threshold on min number of detections
                if detec_tmp >= min_detec_nb:
                    detec_tmp = True
                else:
                    detec_tmp = False
                if annot_tmp >= 1:
                    annot_tmp = True
                else:
                    annot_tmp = False

                # count FP, TP, FN:
                if (annot_tmp is True) & (detec_tmp is True):
                    TP[idx] = 1
                elif (annot_tmp is True) & (detec_tmp is False):
                    FN[idx] = 1
                elif (annot_tmp is False) & (detec_tmp is True):
                    FP[idx] = 1
                elif (annot_tmp is False) & (detec_tmp is False):
                    pass

            # create df for each file with threshold, TP, FP, FN
            tmp = pd.DataFrame(
                {
                    "time": dates_list,
                    "threshold": threshold,
                    "TP": TP,
                    "FP": FP,
                    "FN": FN,
                }
            )
            if th_idx == 0:
                performance_per_interval_count = tmp
            else:
                performance_per_interval_count = pd.concat(
                    [performance_per_interval_count, tmp], ignore_index=True
                )

            # Calculate P, R, and F for that confidence threshold
            perf_tmp = performance_per_interval_count.query(
                "threshold ==" + str(threshold)
            )
            TP_th = perf_tmp["TP"].sum()
            FP_th = perf_tmp["FP"].sum()
            FN_th = perf_tmp["FN"].sum()
            R_th = TP_th / (TP_th + FN_th)
            if (TP_th + FP_th) == 0:
                P_th = 1
            else:
                P_th = TP_th / (TP_th + FP_th)
            F_th = (1 + F_beta**2) * (
                (P_th * R_th) / ((F_beta**2 * P_th) + R_th)
            )
            tmp_PRF = pd.DataFrame(
                {
                    "threshold": [threshold],
                    "TP": [TP_th],
                    "FP": [FP_th],
                    "FN": [FN_th],
                    "R": [R_th],
                    "P": [P_th],
                    "F": [F_th],
                }
            )
            # Sanity check
            # if FP_th + TP_th != len(detec_conf):
            #     raise Exception(
            #         "FP and TP don't add up to the total number of detections"
            #     )
            # elif TP_th + FN_th != len(annot):
            #     raise Exception(
            #         "FP and FN don't add up to the total number of annotations"
            #     )

            if th_idx == 0:
                performance_PRF = tmp_PRF
            else:
                performance_PRF = pd.concat(
                    [performance_PRF, tmp_PRF], ignore_index=True
                )

        # plot PRF curves
        fig, ax = plt.subplots(1, 2)
        # ax[0].axis('equal')
        ax[0].plot(performance_PRF["P"], performance_PRF["R"], "k")
        ax[0].set_xlabel("$Precision\ _{" + integration_time + "}$")
        ax[0].set_ylabel("$Recall\ _{" + integration_time + "}$")
        ax[0].set(xlim=(0, 1), ylim=(0, 1))
        ax[0].grid()
        ax[0].set_aspect("equal", "box")
        fig.tight_layout()

        # ax[1].axis('equal')
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["P"],
            ":k",
            label="$Precision\ _{" + integration_time + "}$",
        )
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["R"],
            "--k",
            label="$Recall\ _{" + integration_time + "}$",
        )
        ax[1].plot(
            performance_PRF["threshold"],
            performance_PRF["F"],
            "k",
            label="$F_" + str(F_beta) + " score\ _{" + integration_time + "}$",
        )
        ax[1].set_xlabel("Threshold")
        ax[1].set_ylabel("Score")
        ax[1].set(xlim=[0, 1], ylim=[0, 1])
        ax[1].grid()
        ax[1].legend()
        ax[1].set_aspect("equal", "box")
        fig.set_size_inches(8, 4)
        fig.tight_layout()
        fig.savefig(
            os.path.join(
                out_dir, "Performance_" + integration_time + "_graph.png"
            )
        )

        # save CSV files
        performance_PRF.to_csv(
            os.path.join(
                out_dir,
                "Performance_" + integration_time + "_full_dataset.csv",
            ),
            index=False,
        )
        performance_per_interval_count.to_csv(
            os.path.join(
                out_dir, "Performance_per_" + integration_time + ".csv"
            ),
            index=False,
        )

    @staticmethod
    def _plot_annot_boxes(
        annot_list,
        out_dir,
        threshold,
        file,
        line_width=2,
        colors=["blue"],
        labels=["Annotations"],
        title="",
    ):
        fig, ax = plt.subplots()
        facecolor = "none"
        alpha = 1
        for idx, annot in enumerate(annot_list):
            patches = []
            for index, row in annot.data.iterrows():
                # plot annotation boxes in Time-frquency
                x = row["time_min_offset"]
                y = row["frequency_min"]
                width = row["duration"]
                height = row["frequency_max"] - row["frequency_min"]
                rect = Rectangle(
                    (x, y),
                    width,
                    height,
                    linewidth=line_width,
                    edgecolor=colors[idx],
                    facecolor=facecolor,
                    alpha=alpha,
                    label=labels[idx],
                )
                patches.append(rect)
            p = PatchCollection(
                patches,
                edgecolor=colors[idx],
                label=labels[idx],
                facecolor=facecolor,
            )
            ax.add_collection(p)
            # ax.add_patch(rect)
        ax.set_xlabel("Times (s)")
        ax.set_ylabel("Frequency (Hz)")
        ax.set_title(title)
        #ax.legend(handles=p)
        ax.grid()
        ax.plot([1], [1])
        #plt.show()
        fig.savefig(os.path.join(out_dir, str(round(threshold,2)) + "_" + file + '.png'))
        plt.close(fig)



